# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

"""Show how to use DDP, Horovod and DeepSpeed strategies interchangeably
with a simple neural network trained on MNIST dataset.
"""

import argparse
import sys
import time
from timeit import default_timer as timer
from typing import Tuple

import deepspeed
import horovod.torch as hvd
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import Dataset
from torchvision import datasets, transforms

from itwinai.parser import ArgumentParser as ItwinaiArgParser
from itwinai.torch.distributed import (
    DeepSpeedStrategy,
    HorovodStrategy,
    NonDistributedStrategy,
    TorchDDPStrategy,
    TorchDistributedStrategy,
    distributed_resources_available,
)
from itwinai.torch.reproducibility import seed_worker, set_seed


def parse_params() -> argparse.Namespace:
    """
    Parse CLI args, which can also be loaded from a configuration file
    using the --config flag:

    >>> train.py --strategy ddp --config config.yaml
    """
    parser = ItwinaiArgParser(description="PyTorch MNIST Example")

    # Distributed ML strategy
    parser.add_argument(
        "--strategy",
        "-s",
        type=str,
        choices=["ddp", "horovod", "deepspeed"],
        default="ddp",
    )

    # Data and logging
    parser.add_argument(
        "--data-dir",
        default="./",
        help=("location of the training dataset in the local " "filesystem"),
    )
    parser.add_argument(
        "--log-int", type=int, default=10, help="log interval per training"
    )
    parser.add_argument(
        "--verbose",
        action=argparse.BooleanOptionalAction,
        help="Print parsed arguments",
    )
    parser.add_argument(
        "--restart-int",
        type=int,
        default=10,
        help="restart interval per epoch (default: 10)",
    )
    parser.add_argument(
        "--download-only",
        action=argparse.BooleanOptionalAction,
        help="Download dataset and exit",
    )
    parser.add_argument(
        "--dataset-replication",
        type=int,
        default=100,
        help="concatenate MNIST to this factor (default: 100)",
    )
    parser.add_argument(
        "--shuff",
        action="store_true",
        default=False,
        help="shuffle dataset (default: False)",
    )
    parser.add_argument(
        "--nworker",
        type=int,
        default=0,
        help=("number of workers in DataLoader (default: 0 -" " only main)"),
    )
    parser.add_argument(
        "--prefetch",
        type=int,
        default=2,
        help="prefetch data in DataLoader (default: 2)",
    )

    # Model
    parser.add_argument(
        "--batch-size",
        type=int,
        default=64,
        help="input batch size for training (default: 64)",
    )
    parser.add_argument(
        "--epochs", type=int, default=10, help="number of epochs to train (default: 10)"
    )
    parser.add_argument(
        "--lr", type=float, default=0.01, help="learning rate (default: 0.01)"
    )
    parser.add_argument(
        "--momentum",
        type=float,
        default=0.5,
        help="momentum in SGD optimizer (default: 0.5)",
    )

    # Reproducibility
    parser.add_argument(
        "--rnd-seed",
        type=int,
        default=0,
        help="seed integer for reproducibility (default: 0)",
    )

    # Distributed ML
    parser.add_argument(
        "--backend",
        type=str,
        default="nccl",
        help="backend for parrallelisation (default: nccl)",
    )
    parser.add_argument(
        "--local_rank",
        type=int,
        default=-1,
        help="local rank passed from distributed launcher",
    )

    # Horovod: ignored when not using Horovod
    parser.add_argument(
        "--fp16-allreduce",
        action="store_true",
        default=False,
        help="use fp16 compression during allreduce",
    )
    parser.add_argument(
        "--use-adasum",
        action="store_true",
        default=False,
        help="use adasum algorithm to do reduction",
    )
    parser.add_argument(
        "--gradient-predivide-factor",
        type=float,
        default=1.0,
        help=("apply gradient pre-divide factor in optimizer " "(default: 1.0)"),
    )

    # DeepSpeed
    parser = deepspeed.add_config_arguments(parser)
    args = parser.parse_args()

    if args.verbose:
        args_list = [f"{key}: {val}" for key, val in args.items()]
        print("PARSED ARGS:\n", "\n".join(args_list))

    return args


class Net(nn.Module):
    """
    Simple neural network classifier for MNIST images.
    """

    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, 10)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return F.log_softmax(x, dim=-1)


def train(
    model, train_loader, optimizer, epoch, strategy: TorchDistributedStrategy, args
):
    """
    Training function, representing an epoch.
    """
    model.train()
    t_list = []
    loss_acc = 0
    if strategy.is_main_worker:
        print("\n")
    for batch_idx, (data, target) in enumerate(train_loader):
        t = timer()
        data = data.to(strategy.device())
        target = target.to(strategy.device())
        optimizer.zero_grad()
        output = model(data)
        loss = F.nll_loss(output, target)
        loss.backward()
        optimizer.step()
        if (
            strategy.is_main_worker
            and args.log_int > 0
            and batch_idx % args.log_int == 0
        ):
            dl_size = len(train_loader.dataset) // strategy.global_world_size()
            print(
                f"Train epoch: {epoch} "
                f"[{batch_idx * len(data)}/{dl_size} "
                f"({100.0 * batch_idx / len(train_loader):.0f}%)]\t\t"
                f"Loss: {loss.item():.6f}"
            )
        t_list.append(timer() - t)
        loss_acc += loss.item()
    if strategy.is_main_worker:
        print("TIMER: train time", sum(t_list) / len(t_list), "s")
    return loss_acc


def test(model, test_loader, strategy: TorchDistributedStrategy):
    """
    Model validation.
    """
    model.eval()
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data = data.to(strategy.device())
            target = target.to(strategy.device())
            output = model(data)
            # Sum up batch loss
            test_loss += F.nll_loss(output, target, reduction="sum").item()
            # Get the index of the max log-probability
            pred = output.argmax(dim=1, keepdim=True)
            correct += pred.eq(target.view_as(pred)).sum().item()
    test_loss /= len(test_loader.dataset)
    if strategy.is_main_worker:
        dl_size = len(test_loader.dataset) // strategy.global_world_size()
        print(
            f"Test set: average loss: {test_loss:.4f}\t"
            f"accurate samples: {correct}/{dl_size}"
        )
    acc_test = 100.0 * correct * strategy.global_world_size() / len(test_loader.dataset)
    return acc_test


def download_mnist():
    """
    Use built-in torch datasets functions to pull MNIST dataset.
    """

    _ = datasets.MNIST(
        args.data_dir,
        train=True,
        download=True,
        transform=transforms.Compose(
            [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
        ),
    )
    _ = datasets.MNIST(
        args.data_dir,
        train=False,
        download=True,
        transform=transforms.Compose(
            [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
        ),
    )


def mnist_dataset(dataset_replication: int = 1) -> Tuple[Dataset, Dataset]:
    """Load MNIST train and test datasets, replicating them.

    Args:
        dataset_replication (int): dataset replication factor. Default 1.

    Returns:
        Tuple[Dataset, Dataset]: train dataset and test dataset.
    """
    replicated_data = [
        datasets.MNIST(
            args.data_dir,
            train=True,
            download=False,
            transform=transforms.Compose(
                [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
            ),
        )
        for _ in range(dataset_replication)
    ]
    train_dataset = torch.utils.data.ConcatDataset(replicated_data)

    replicated_data = [
        datasets.MNIST(
            args.data_dir,
            train=False,
            download=False,
            transform=transforms.Compose(
                [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
            ),
        )
        for _ in range(dataset_replication)
    ]
    test_dataset = torch.utils.data.ConcatDataset(replicated_data)
    return train_dataset, test_dataset


if __name__ == "__main__":
    args = parse_params()

    if args.download_only:
        # Download datasets from a location with internet access and exit.
        # This is convenient when submitting training jobs to
        # a batch system where worker nodes have no internet
        # access, like in some HPCs.
        download_mnist()
        sys.exit()

    # Instantiate Strategy
    if not distributed_resources_available():
        print("WARNING: falling back to non-distributed strategy.")
        strategy = NonDistributedStrategy()
        distribute_kwargs = {}
    elif args.strategy == "ddp":
        strategy = TorchDDPStrategy(backend=args.backend)
        distribute_kwargs = {}
    elif args.strategy == "horovod":
        strategy = HorovodStrategy()
        distribute_kwargs = dict(
            compression=(
                hvd.Compression.fp16 if args.fp16_allreduce else hvd.Compression.none
            ),
            op=hvd.Adasum if args.use_adasum else hvd.Average,
            gradient_predivide_factor=args.gradient_predivide_factor,
        )
    elif args.strategy == "deepspeed":
        strategy = DeepSpeedStrategy(backend=args.backend)
        distribute_kwargs = dict(
            config_params=dict(train_micro_batch_size_per_gpu=args.batch_size)
        )
    else:
        raise NotImplementedError(
            f"Strategy {args.strategy} is not recognized/implemented."
        )

    # Initialize strategy
    strategy.init()

    # Start the timer for profiling
    st = timer()

    # Set random seed for reproducibility
    torch_prng = set_seed(args.rnd_seed)

    if strategy.is_main_worker:
        print("TIMER: initialise:", timer() - st, "s")
        print(
            "DEBUG: local ranks:",
            strategy.local_world_size(),
            "/ global ranks:",
            strategy.global_world_size(),
        )
        print("DEBUG: sys.version:", sys.version)
        print("DEBUG: args.data_dir:", args.data_dir)
        print("DEBUG: args.log_int:", args.log_int)
        print("DEBUG: args.nworker:", args.nworker)
        print("DEBUG: args.prefetch:", args.prefetch)
        print("DEBUG: args.batch_size:", args.batch_size)
        print("DEBUG: args.epochs:", args.epochs)
        print("DEBUG: args.lr:", args.lr)
        print("DEBUG: args.momentum:", args.momentum)
        print("DEBUG: args.shuff:", args.shuff)
        print("DEBUG: args.rnd_seed:", args.rnd_seed)
        print("DEBUG: args.backend:", args.backend)

    # Dataset
    train_dataset, test_dataset = mnist_dataset(args.dataset_replication)
    # Distributed dataloaders
    train_loader = strategy.create_dataloader(
        train_dataset,
        batch_size=args.batch_size,
        num_workers=args.nworker,
        pin_memory=True,
        persistent_workers=(args.nworker > 1),
        prefetch_factor=args.prefetch,
        generator=torch_prng,
        worker_init_fn=seed_worker,
    )
    test_loader = strategy.create_dataloader(
        test_dataset,
        batch_size=args.batch_size,
        num_workers=args.nworker,
        pin_memory=True,
        persistent_workers=(args.nworker > 1),
        prefetch_factor=args.prefetch,
        generator=torch_prng,
        worker_init_fn=seed_worker,
    )

    if strategy.is_main_worker:
        print("TIMER: read and concat data:", timer() - st, "s")

    # Create CNN model
    model = Net().to(strategy.device())

    # Optimizer
    optimizer = torch.optim.SGD(model.parameters(), lr=args.lr, momentum=args.momentum)

    # Distributed model, optimizer, and scheduler
    model, optimizer, _ = strategy.distributed(
        model, optimizer, lr_scheduler=None, **distribute_kwargs
    )

    # Start training and test loop
    if strategy.is_main_worker:
        print("TIMER: broadcast:", timer() - st, "s")
        print("\nDEBUG: start training")
        print("--------------------------------------------------------")

    et = timer()
    start_epoch = 1
    for epoch in range(start_epoch, args.epochs + 1):
        lt = timer()
        if strategy.is_distributed:
            # Inform the sampler that a new epoch started: shuffle
            # may be needed
            train_loader.sampler.set_epoch(epoch)
            test_loader.sampler.set_epoch(epoch)

        # Training
        loss_acc = train(
            model=model,
            train_loader=train_loader,
            optimizer=optimizer,
            epoch=epoch,
            strategy=strategy,
            args=args,
        )

        # Testing
        acc_test = test(model=model, test_loader=test_loader, strategy=strategy)

        # Save first epoch timer
        if epoch == start_epoch:
            first_ep_t = timer() - lt

        # Final epoch
        if epoch + 1 == args.epochs:
            train_loader.last_epoch = True
            test_loader.last_epoch = True

        if strategy.is_main_worker:
            print("TIMER: epoch time:", timer() - lt, "s")
            print("DEBUG: accuracy:", acc_test, "%")

    if strategy.is_main_worker:
        print("\n--------------------------------------------------------")
        print("DEBUG: training results:\n")
        print("TIMER: first epoch time:", first_ep_t, " s")
        print("TIMER: last epoch time:", timer() - lt, " s")
        print("TIMER: average epoch time:", (timer() - et) / args.epochs, " s")
        print("TIMER: total epoch time:", timer() - et, " s")
        if epoch > 1:
            print("TIMER: total epoch-1 time:", timer() - et - first_ep_t, " s")
            print(
                "TIMER: average epoch-1 time:",
                (timer() - et - first_ep_t) / (args.epochs - 1),
                " s",
            )
        print("DEBUG: last accuracy:", acc_test, "%")
        if torch.cuda.is_available():
            print(
                "DEBUG: memory req:",
                int(torch.cuda.memory_reserved(strategy.local_rank()) / 1024 / 1024),
                "MB",
            )
            print("DEBUG: memory summary:\n\n", torch.cuda.memory_summary(0))

        print(f"TIMER: final time: {timer()-st} s\n")

    time.sleep(1)
    print(f"<Global rank: {strategy.global_rank()}> - TRAINING FINISHED")

    # Clean-up
    strategy.clean_up()
    sys.exit()



# This script trains the Convolutional Variational Auto-Encoder (CVAE)
# network on preprocessed CMIP6 Data


"""
from codecarbon import EmissionsTracker

# Instantiate the tracker object
tracker = EmissionsTracker(
    output_dir="../code_carbon/",  # define the directory where to write the emissions results
    output_file="emissions.csv",  # define the name of the file containing the emissions results
    # log_level='error' # comment out this line to see regular output
)
tracker.start()
"""

from typing import Any, Dict, List, Literal, Optional, Tuple
import torch
import torch.optim as optim
import numpy as np
import pandas as pd
from pathlib import Path
import sys
import os

from itwinai.components import monitor_exec
from itwinai.torch.trainer import TorchTrainer
from itwinai.torch.distributed import DeepSpeedStrategy
from itwinai.loggers import Logger
from itwinai.torch.config import TrainingConfiguration

from ray import train as train_ray
import src.model as model
from torch.utils.data import Dataset, DataLoader
from torchvision.utils import make_grid
from src.engine import train as train_engine
from src.engine import validate as validate_engine
from src.utils import save_reconstructed_images, save_loss_plot, save_ex
from src.initialization import beta, criterion, pixel_wise_criterion

class XTClimTrainer(TorchTrainer):
    def __init__(
            self,
            epochs: int,
            config: Dict | TrainingConfiguration | None = None,
            seasons: Literal["winter_", "spring_", "summer_", "autumn_"] = 'winter_',
            strategy: Literal["ddp", "deepspeed", "horovod"] = 'ddp',
            save_best: bool = True,
            logger: Logger | None = None
    ):
        super().__init__(
            epochs=epochs,
            config=config,
            strategy=strategy,
            logger=logger
        )
        self.epochs = epochs
        self.seasons = seasons
        # Global training configuration
        if isinstance(config, dict):
            config = TrainingConfiguration(**config)
        self.config = config

    def final_loss(
        self,
        bce_loss: torch.Tensor,
        mu: torch.Tensor,
        logvar: torch.Tensor,
        beta: float = 0.1
    ) -> torch.Tensor:
        """Adds up reconstruction loss (BCELoss) and Kullback-Leibler divergence.
        KL-Divergence = 0.5 * sum(1 + log(sigma^2) - mu^2 - sigma^2)

        Args:
            bce_loss (torch.Tensor): recontruction loss
            mu (torch.Tensor): the mean from the latent vector
            logvar (torch.Tensor): log variance from the latent vector
            beta (torch.Tensor): weight over the KL-Divergence

        Returns:
            total loss (torch.Tensor)
        """
        KLD = -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp())
        return bce_loss + beta*KLD

    def create_dataloaders(
        self,
        train_dataset: Dataset,
        validation_dataset: Dataset | None = None,
        test_dataset: Dataset | None = None
    ) -> None: #Tuple[DataLoader, DataLoader, List, List]:
        """Create the training and test dataloaders.

        Args:
            train_dataset (Dataset): training dataset object.
            validation_dataset (Dataset | None): validation dataset object.
                Default None.
            test_dataset (Dataset | None): test dataset object.
                Default None.

        Returns:
            Tuple[DataLoader, DataLoader, List, List]: tuple containing dataloaders and samples.
        """

        # Create dataloaders
        self.trainloader = self.strategy.create_dataloader(
            self.train_dataset, batch_size=self.config.batch_size, shuffle=True, pin_memory=True
        )
        self.testloader = self.strategy.create_dataloader(
            self.test_dataset, batch_size=self.config.batch_size, shuffle=False, pin_memory=True
        )

    def _handle_early_stopping(self, epoch: int, valid_epoch_loss: float) -> bool:
        """Handle early stopping based on validation loss.

        Args:
            epoch (int): Current epoch number.
            valid_epoch_loss (float): Validation loss for current epoch.

        Returns:
            bool: True if training should stop early, False otherwise.
        """
        min_epochs = epoch > 1
        improve_below_threshold = (
            (self.config.old_valid_loss - valid_epoch_loss) / self.config.old_valid_loss
        ) < self.config.stop_delta
        if min_epochs and improve_below_threshold:
            # if the marginal improvement in validation loss is too small
            self.config.early_count += 1
            if self.config.early_count > self.config.patience:
                # if too small improvement for a few epochs in a row, stop learning
                print(f"Early stopping at epoch {epoch}")
                save_ex(recon_images[0], epoch, self.seasons)
                return True
        else:
            # if the condition is not verified anymore, reset the count
            self.config.early_count = 0

    def _save_best_model(self, valid_epoch_loss: float, epoch: int) -> None:
        """Save the model checkpoint if best.

        Args:
            valid_epoch_loss (float): Validation loss for current epoch.
            epoch (int): Current epoch number.

        Returns:
            None
        """
        worker_val_losses = self.strategy.gather_obj(valid_epoch_loss)
        if self.strategy.is_main_worker:
            # Save only in the main worker
            # avg_loss has a meaning only in the main worker
            avg_loss = np.mean(worker_val_losses)
            if self.config.save_best and avg_loss < self.min_valid_epoch_loss:
                self.min_valid_epoch_loss = avg_loss
                checkpoint = {
                    'epoch': epoch,
                    'model_state_dict': self.model.state_dict(),
                    'optim_state_dict': self.optimizer.state_dict(),
                    'val_loss': valid_epoch_loss,
                }
                # save checkpoint only if it is better than the previous ones
                checkpoint_filename = self.output_dir / f"cvae_model_{self.seasons}1d_{self.config.n_memb}memb.pth"
                torch.save(checkpoint, checkpoint_filename)
                # itwinai - log checkpoint as artifact
                self.log(checkpoint_filename, Path(checkpoint_filename).name, kind='artifact')

    def create_model_loss_optimizer(self) -> None:
        # initialize the model
        cvae_model = model.ConvVAE()
        optimizer = optim.Adam(cvae_model.parameters(), lr=self.config.optim_lr)

        # First, define strategy-wise optional configurations
        if isinstance(self.strategy, DeepSpeedStrategy):
            # Batch size definition is not optional for DeepSpeedStrategy!
            distribute_kwargs = dict(
                config_params=dict(
                    train_micro_batch_size_per_gpu=self.config.batch_size
                )
            )
        else:
            distribute_kwargs = {}

        # Distributed model, optimizer, and scheduler
        self.model, self.optimizer, _ = self.strategy.distributed(
            cvae_model, optimizer, **distribute_kwargs
        )

    def init_dataloading_step(self):
        self.input_dir = Path("input")
        self.output_dir = Path("outputs")

        # Load train data
        train_time = pd.read_csv(
            self.input_dir / f"dates_train_{self.seasons}data_{self.config.n_memb}memb.csv"
        )
        train_data = np.load(
            self.input_dir / f"preprocessed_1d_train_{self.seasons}data_{self.config.n_memb}memb.npy"
        )
        n_train = len(train_data)
        train_dataset = [
            (torch.from_numpy(np.reshape(train_data[i], (2, 32, 32))), train_time["0"][i])
            for i in range(n_train)
        ]

        # Load test data
        test_time = pd.read_csv(
            self.input_dir / f"dates_test_{self.seasons}data_{self.config.n_memb}memb.csv"
        )
        test_data = np.load(
            self.input_dir / f"preprocessed_1d_test_{self.seasons}data_{self.config.n_memb}memb.npy"
        )
        n_test = len(test_data)
        test_dataset = [
            (torch.from_numpy(np.reshape(test_data[i], (2, 32, 32))), test_time["0"][i])
            for i in range(n_test)
        ]

        return train_dataset, test_dataset

    def train(self):
        """Trains the XTClim model."""

        # Initialize lists to track loss and images
        grid_images = []
        train_loss = []
        valid_loss = []
        self.min_valid_epoch_loss = float('inf')  # random initial value

        for epoch in range(self.epochs):
            if self.strategy.is_main_worker:
                print(f"Epoch {epoch+1} of {self.epochs}")

            if self.strategy.is_distributed:
                # Inform the sampler that a new epoch started: shuffle may be needed
                self.trainloader.sampler.set_epoch(epoch)
                self.testloader.sampler.set_epoch(epoch)

            # train the model
            train_epoch_loss = train_engine(
                self.model, self.trainloader, self.train_dataset, self.device, self.optimizer, criterion, beta
            )

            # evaluate the model on the test set
            valid_epoch_loss, recon_images = validate_engine(
                self.model, self.testloader, self.test_dataset, self.device, criterion, beta
            )

            self.log(train_epoch_loss, 'epoch_train_loss', kind='metric')
            self.log(valid_epoch_loss, 'epoch_valid_loss', kind='metric')

            # keep track of the losses
            train_loss.append(train_epoch_loss)
            valid_loss.append(valid_epoch_loss)

            # Convert the reconstructed images to PyTorch grid format
            image_grid = make_grid(recon_images.detach().cpu())
            grid_images.append(image_grid)

            # decreasing learning rate
            if (epoch + 1) % self.config.lr_decay_interval == 0:
                self.config.optim_lr = self.config.optim_lr / self.config.lr_decay_rate

            # Early stopping and learning rate decay
            if self._handle_early_stopping(epoch, valid_epoch_loss):
                break
            self.config.old_valid_loss = valid_epoch_loss

            # Save the best model checkpoint
            self._save_best_model(valid_epoch_loss, epoch)

        # Final loss reports
        print(f"Train Loss: {train_epoch_loss:.4f}")
        print(f"Val Loss: {valid_epoch_loss:.4f}")

        # Report training metrics of last epoch to Ray
        train_ray.report(
            {
                "loss": train_epoch_loss,
                "valid_loss": valid_epoch_loss
            }
        )

        save_loss_plot(train_loss, valid_loss, self.seasons)
        # save the loss evolutions
        pd.DataFrame(train_loss).to_csv(
            self.output_dir / f"train_loss_indiv_{self.seasons}1d_{self.config.n_memb}memb.csv"
        )
        pd.DataFrame(valid_loss).to_csv(
            self.output_dir / f"test_loss_indiv_{self.seasons}1d_{self.config.n_memb}memb.csv"
        )

    @monitor_exec
    def execute(self) -> None:
        self.train_dataset, self.test_dataset = self.init_dataloading_step()

        _ = super().execute(self.train_dataset, self.test_dataset)

        return None


# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Linus Eickhoff
#
# Credit:
# - Linus Eickhoff <linus.maximilian.eickhoff@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import argparse
import re
from pathlib import Path
from typing import List


def convert_github_admonitions(content: str) -> str:
    """Converts GitHub-style admonitions to sphinx md syntax.

    Args:
        content (str): The markdown content containing GitHub-style admonitions

    Returns:
        str: The markdown content with GitHub admonitions converted to sphinx md syntax

    Example:
    > [!NOTE]
    > This is a note

    Becomes:
    ```{note}
    This is a note
    ```
    """
    # Github markdown pattern
    pattern = r">\s*\[!(NOTE|WARNING|TIP|IMPORTANT|CAUTION)\][ \t]*(.*(?:\n>.*)*)"

    def replace_admonition(match):
        # same name in lowercase for sphinx-compatible admonitions
        admonition_type = match.group(1).lower()
        content_lines = match.group(2)

        cleaned_content = []
        for line in content_lines.splitlines():
            if line.strip():
                # Preserve indentation after removing > prefix
                cleaned_line = re.sub(r"^>\s?", "", line)
                cleaned_content.append(cleaned_line)

        return "```{" + f"{admonition_type}" + "}\n" + "\n".join(cleaned_content) + "\n```"

    return re.sub(pattern, replace_admonition, content)


def find_md_includes_in_rst(rst_content: str, rst_file_path: str | Path) -> List[Path]:
    """Find all .md file includes in an RST file and return their absolute paths.

    Args:
        rst_content (str): Content of the RST file to search for includes
        rst_file_path (str): Path to the RST file, used for resolving relative paths

    Returns:
        list[Path]: List of absolute Path objects pointing to included .md files
    """
    pattern = r"\.\.\s+include::\s+(.+\.md)"
    matches = re.finditer(pattern, rst_content)
    md_files = []
    rst_path = Path(rst_file_path)

    for match in matches:
        md_path = match.group(1)
        # Handle relative paths properly using Path
        abs_md_path = (rst_path.parent / md_path).resolve()
        md_files.append(abs_md_path)

    return md_files


def process_rst_files(docs_dir: str) -> None:
    """Process all RST files in the docs directory and convert referenced MD files.

    Args:
        docs_dir (str): Path to the docs directory containing RST files

    Returns:
        None: This function processes files in-place and prints status messages
    """
    docs_path = Path(docs_dir).resolve()
    processed_files = set()  # To prevent processing the same file multiple times

    for rst_file in docs_path.rglob("*.rst"):
        if rst_file.is_symlink():  # Skip symlinks to prevent loops
            print(f"Skipping symlink: {rst_file}")
            continue

        print(f"Processing RST file: {rst_file}")

        try:
            # Read RST content
            rst_content = rst_file.read_text(encoding="utf-8")
            # Find MD includes
            md_files = find_md_includes_in_rst(rst_content, rst_file)
            # Process each MD file
            for md_file in md_files:
                try:
                    if md_file in processed_files:
                        continue

                    processed_files.add(md_file)

                    if not md_file.exists():
                        print(f"Warning: Referenced MD file not found: {md_file}")
                        continue

                    print(f"Checking for admonitions in: {md_file}")
                    # assume utf8
                    md_content = md_file.read_text(encoding="utf-8")
                    # Convert admonitions
                    converted_content = convert_github_admonitions(md_content)
                    # Print to stdout where changes were made
                    if converted_content != md_content:
                        md_file.write_text(converted_content, encoding="utf-8")
                        print(f"Changed admonitions in: {md_file}")

                except Exception as e:
                    print(f"Unexpected error processing {md_file}: {str(e)}")

        except Exception as e:
            print(f"Unexpected error processing {rst_file}: {str(e)}")


if __name__ == "__main__":
    script_dir = Path(__file__).parent.resolve()

    parser = argparse.ArgumentParser(
        description="Convert GitHub-style admonitions of md references to sphinx syntax"
    )
    parser.add_argument(
        "-d",
        "--dir",
        type=str,
        required=True,
        help="Path to directory where to search for RST files",
    )
    args = parser.parse_args()

    process_rst_files(args.dir)



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Jarl Sondre Sæther
#
# Credit:
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# --------------------------------------------------------------------------------------

from typing import List

from itwinai.parser import ArgumentParser


def remove_indentation_from_multiline_string(multiline_string: str) -> str:
    """Removes the indentation from the start of each line in a multi-line string. The
    main purpose of this function is allowing you to define multi-line strings that
    don't touch the left margin of the editor, thus increasing readability.
    """
    return "\n".join([line.lstrip() for line in multiline_string.split("\n")])


def scalability_nodes_list(value: str | List[int]) -> List[int]:
    """Checks that the value it receives conforms to the comma-separated integer
    constraint and returns the parsed list if successful.

    Returns:
        The list of integers that was parsed.

    Raises:
        ValueError: If unable to parse the integers e.g. due to formatting errors.
    """

    if isinstance(value, list):
        if not all([isinstance(x, int) for x in value]):
            raise ValueError(f"Provided list, '{value}', contains non-integer values.")
        else:
            return value

    try:
        return [int(n) for n in value.split(",")]
    except ValueError:
        raise ValueError(
            f"Invalid input: '{value}', must be formatted as comma-separated integers."
        )


def get_slurm_job_parser() -> ArgumentParser:
    # Default arguments for the SLURM script configuration
    default_account = "intertwin"
    default_time = "00:30:00"
    default_partition = "develbooster"
    default_job_name = None
    default_std_out = None
    default_err_out = None
    default_num_nodes = 1
    default_num_tasks_per_node = 1
    default_gpus_per_node = 4
    default_cpus_per_gpu = 4

    # Default other arguments
    default_mode = "single"
    default_distributed_strategy = "ddp"
    default_config_file_path = "."
    default_config_file = "config"
    default_pipe_key = "rnn_training_pipeline"
    # Command to be executed before the main process starts.
    default_pre_exec_command = None
    default_training_command = None
    default_python_venv = ".venv"
    default_scalability_nodes = "1,2,4,8"
    default_profiling_rate = 10 # py-spy profiler sample rate/frequency

    parser = ArgumentParser(parser_mode="omegaconf")

    # Arguments specific to the SLURM script configuration
    parser.add_argument(
        "--job-name",
        type=str,
        default=default_job_name,
        help="The name of the SLURM job.",
    )
    parser.add_argument(
        "--account",
        type=str,
        default=default_account,
        help="The billing account for the SLURM job.",
    )
    parser.add_argument(
        "--time",
        type=str,
        default=default_time,
        help="The time limit of the SLURM job.",
    )
    parser.add_argument(
        "--partition",
        type=str,
        default=default_partition,
        help="Which partition of the cluster the SLURM job is going to run on.",
    )
    parser.add_argument(
        "--std-out", type=str, default=default_std_out, help="The standard out file."
    )
    parser.add_argument(
        "--err-out", type=str, default=default_err_out, help="The error out file."
    )
    parser.add_argument(
        "--num-nodes",
        type=int,
        default=default_num_nodes,
        help="The number of nodes that the SLURM job is going to run on.",
    )
    parser.add_argument(
        "--num-tasks-per-node",
        type=int,
        default=default_num_tasks_per_node,
        help="The number of tasks per node.",
    )
    parser.add_argument(
        "--gpus-per-node",
        type=int,
        default=default_gpus_per_node,
        help="The requested number of GPUs per node.",
    )
    parser.add_argument(
        "--cpus-per-gpu",
        type=int,
        default=default_cpus_per_gpu,
        help="The requested number of CPUs per node.",
    )

    # Arguments specific to the itwinai pipeline
    parser.add_argument(
        "--config-path",
        type=str,
        default=default_config_file_path,
        help="The path to the directory containing the config file to use for training.",
    )
    parser.add_argument(
        "--config-name",
        type=str,
        default=default_config_file,
        help="The name of the config file to use for training.",
    )
    parser.add_argument(
        "--pipe-key",
        type=str,
        default=default_pipe_key,
        help="Which pipe key to use for running the pipeline.",
    )

    # Arguments specific to the SLURM Builder
    parser.add_argument(
        "--mode",
        choices=["scaling-test", "runall", "single"],
        default=default_mode,
        help="Which mode to run, e.g. scaling test, all strategies, or a single run.",
    )
    parser.add_argument(
        "--dist-strat",
        choices=["ddp", "horovod", "deepspeed"],
        default=default_distributed_strategy,
        help="Which distributed strategy to use.",
    )
    parser.add_argument(
        "--pre-exec-cmd",
        type=str,
        default=default_pre_exec_command,
        help="The pre-execution command to use for the python script.",
    )
    parser.add_argument(
        "--training-cmd",
        type=str,
        default=default_training_command,
        help="The training command to use for the python script.",
    )
    parser.add_argument(
        "--python-venv",
        type=str,
        default=default_python_venv,
        help="Which python venv to use for running the command.",
    )
    parser.add_argument(
        "--scalability-nodes",
        type=scalability_nodes_list,
        default=default_scalability_nodes,
        help="A comma-separated list of node numbers to use for the scalability test.",
    )
    parser.add_argument(
        "-pr",
        "--profiling-rate",
        type=int,
        default=default_profiling_rate,
        help="The rate at which the py-spy profiler should sample the call stack.",
    )

    # Boolean arguments where you only need to include the flag and not an actual value
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Whether to include debugging information or not.",
    )
    parser.add_argument(
        "-ns",
        "--no-save-script",
        action="store_true",
        help="Whether to save the script after processing it.",
    )
    parser.add_argument(
        "-nj",
        "--no-submit-job",
        action="store_true",
        help="Whether to submit the job when processing the script.",
    )
    parser.add_argument(
        "-ps",
        "--py-spy",
        action="store_true",
        help="Whether to activate profiling with py-spy.",
    )

    return parser



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Roman Machacek
#
# Credit:
# - Roman Machacek <roman.machacek@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------


# import tensorflow.keras as keras
from typing import List

import tensorflow as tf


class MNIST_Model(tf.keras.Model):
    def __init__(self, input_shape: List[int] = (28, 28, 1), output_shape: int = 10):
        super().__init__()

        # LeNet5
        self.model = tf.keras.Sequential(
            [
                tf.keras.layers.Conv2D(
                    filters=6, kernel_size=(3, 3), activation="relu", input_shape=(28, 28, 1)
                ),
                tf.keras.layers.AveragePooling2D(2),
                tf.keras.layers.Conv2D(filters=16, kernel_size=(3, 3), activation="relu"),
                tf.keras.layers.AveragePooling2D(2),
                tf.keras.layers.Flatten(),
                tf.keras.layers.Dense(units=120, activation="relu"),
                tf.keras.layers.Dense(units=84, activation="relu"),
                tf.keras.layers.Dense(units=10),
            ]
        )

    def call(self, inputs):
        return self.model(inputs)



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import pytest

pytest.PIPE_LIST_YAML = """
my-list-pipeline:
  _target_: itwinai.pipeline.Pipeline
  steps:
    - _target_: itwinai.tests.dummy_components.FakePreproc
      max_items: 33
      name: my-preproc

    - _target_: itwinai.tests.dummy_components.FakeTrainer
      lr: 0.001
      batch_size: 32
      name: my-trainer
"""

pytest.PIPE_DICT_YAML = """
my-dict-pipeline:
  _target_: itwinai.pipeline.Pipeline
  steps:
    preproc-step:
      _target_: itwinai.tests.dummy_components.FakePreproc
      max_items: 33
      name: my-preproc

    train-step:
      _target_: itwinai.tests.dummy_components.FakeTrainer
      lr: 0.001
      batch_size: 32
      name: my-trainer
"""

pytest.NESTED_PIPELINE = """
some:
  field:
    my-nested-pipeline:
      _target_: itwinai.pipeline.Pipeline
      steps:
        - _target_: itwinai.tests.dummy_components.FakePreproc
          max_items: 33
          name: my-preproc

        - _target_: itwinai.tests.dummy_components.FakeTrainer
          lr: 0.001
          batch_size: 32
          name: my-trainer
"""

pytest.INTERPOLATED_VALUES_PIPELINE = """
max_items: 33
name: my-trainer
lr: 0.001
my-interpolation-pipeline:
  _target_: itwinai.pipeline.Pipeline
  steps:
    - _target_: itwinai.tests.dummy_components.FakePreproc
      max_items: ${max_items}
      name: my-preproc

    - _target_: itwinai.tests.dummy_components.FakeTrainer
      lr: ${lr}
      batch_size: 32
      name: ${name}
"""



import tempfile
from pathlib import Path
from unittest.mock import MagicMock, patch

import mlflow
import pytest

from itwinai.loggers import Prov4MLLogger


@pytest.fixture
def logger_instance():
    return Prov4MLLogger()


@pytest.fixture
def mlflow_run():
    with tempfile.TemporaryDirectory() as mlflow_temp_dir:
        mlflow.set_tracking_uri(Path(mlflow_temp_dir).resolve().as_uri())
        experiment_id = mlflow.create_experiment("temporary_experiment")
        mlflow.set_experiment(experiment_id=experiment_id)
        # nested=True is needed
        yield mlflow.start_run(nested=True)
        mlflow.end_run()



def test_create_destroy_logger_context(logger_instance):
    logger_instance.should_log = MagicMock(return_value=True)
    with patch("prov4ml.start_run") as start_run, patch("prov4ml.end_run") as end_run:
        logger_instance.create_logger_context(rank=1)
        logger_instance.destroy_logger_context()
        end_run.assert_called_once()
        start_run.assert_called_once()


def test_create_destroy_logger_context_should_not_log(logger_instance):
    logger_instance.should_log = MagicMock(return_value=False)
    with patch("prov4ml.start_run") as start_run, patch("prov4ml.end_run") as end_run:
        logger_instance.create_logger_context(rank=1)
        logger_instance.destroy_logger_context()
        end_run.assert_not_called()
        start_run.assert_not_called()


def test_logger_context_start_logging(logger_instance):
    logger_instance.should_log = MagicMock(return_value=True)
    with patch("prov4ml.start_run") as start_run, patch("prov4ml.end_run") as end_run:
        with logger_instance.start_logging(rank=1):
            pass
        start_run.assert_called_once()
        end_run.assert_called_once()


def test_create_logger_context_should_log(logger_instance):
    logger_instance.should_log = MagicMock(return_value=True)
    logger_instance.create_logger_context(rank=1)
    with patch("prov4ml.log_metric") as log_metric:
        logger_instance.log(123, "number", kind="metric")
    log_metric.assert_called_once()


def test_create_logger_context_should_not_log(logger_instance):
    logger_instance.should_log = MagicMock(return_value=False)
    logger_instance.create_logger_context(rank=1)
    with patch("prov4ml.log_metric") as log_metric:
        logger_instance.log(123, "number", kind="metric")
    log_metric.assert_not_called()


def test_save_hyperparameters(logger_instance):
    logger_instance.should_log = MagicMock(return_value=True)
    logger_instance.create_logger_context(rank=1)
    params = {"learning_rate": 0.01, "batch_size": 32}
    with patch("prov4ml.log_param") as log_param:
        logger_instance.save_hyperparameters(params)
        log_param.assert_any_call(key="learning_rate", value=0.01)
        log_param.assert_any_call(key="batch_size", value=32)


def test_log_metric(logger_instance):
    logger_instance.should_log = MagicMock(return_value=True)
    logger_instance.create_logger_context(rank=1)

    with patch("prov4ml.log_metric") as log_metric:
        logger_instance.log(item=0.95, identifier="accuracy", kind="metric", step=1)

        log_metric.assert_called_once_with(
            key="accuracy", value=0.95, step=1, context="training"
        )


def test_log_flops_per_batch(logger_instance):
    logger_instance.should_log = MagicMock(return_value=True)
    logger_instance.create_logger_context(rank=1)

    model_mock, batch_mock = MagicMock(), MagicMock()
    with patch("prov4ml.log_flops_per_batch") as log_flops_pb:
        logger_instance.log(
            item=(model_mock, batch_mock), identifier="my_flops_pb", kind="flops_pb", step=1
        )

        log_flops_pb.assert_called_once_with(
            model=model_mock, batch=batch_mock, label="my_flops_pb", step=1, context="training"
        )


def test_log_flops_per_epoch(logger_instance):
    logger_instance.should_log = MagicMock(return_value=True)
    logger_instance.create_logger_context(rank=1)

    model_mock, dataset_mock = MagicMock(), MagicMock()
    with patch("prov4ml.log_flops_per_epoch") as log_flops_pe:
        logger_instance.log(
            item=(model_mock, dataset_mock), identifier="my_flops_pe", kind="flops_pe", step=1
        )

        log_flops_pe.assert_called_once_with(
            model=model_mock,
            dataset=dataset_mock,
            label="my_flops_pe",
            step=1,
            context="training",
        )


def test_log_system(logger_instance):
    logger_instance.should_log = MagicMock(return_value=True)
    logger_instance.create_logger_context(rank=1)

    with patch("prov4ml.log_system_metrics") as log_system:
        logger_instance.log(item=None, identifier=None, kind="system", step=1)

        log_system.assert_called_once_with(context="training", step=1)


def test_log_carbon(logger_instance):
    logger_instance.should_log = MagicMock(return_value=True)
    logger_instance.create_logger_context(rank=1)

    with patch("prov4ml.log_carbon_metrics") as log_carbon:
        logger_instance.log(item=None, identifier=None, kind="carbon", step=1)

        log_carbon.assert_called_once_with(context="training", step=1)


def test_log_execution_time(logger_instance):
    logger_instance.should_log = MagicMock(return_value=True)
    logger_instance.create_logger_context(rank=1)

    with patch("prov4ml.log_current_execution_time") as log_exec_time:
        logger_instance.log(
            item=None, identifier="execution_time", kind="execution_time", step=1
        )

        log_exec_time.assert_called_once_with(
            label="execution_time", context="training", step=1
        )


def test_log_model(logger_instance):
    logger_instance.should_log = MagicMock(return_value=True)
    logger_instance.create_logger_context(rank=1)

    model_mock = MagicMock()
    with patch("prov4ml.save_model_version") as log_model:
        logger_instance.log(item=model_mock, identifier="model_v1", kind="model", step=1)

        log_model.assert_called_once_with(
            model=model_mock, model_name="model_v1", context="training", step=1
        )


def test_log_best_model(logger_instance):
    logger_instance.should_log = MagicMock(return_value=True)
    logger_instance.create_logger_context(rank=1)

    model_mock = MagicMock()
    with patch("prov4ml.log_model") as log_best_model:
        logger_instance.log(
            item=model_mock, identifier="best_model_v1", kind="best_model", step=1
        )

        log_best_model.assert_called_once_with(
            model=model_mock,
            model_name="best_model_v1",
            log_model_info=True,
            log_as_artifact=True,
        )


def test_log_prov_documents(logger_instance, mlflow_run):
    logger_instance.should_log = MagicMock(return_value=True)
    logger_instance.create_logger_context(rank=1)

    with patch("prov4ml.log_provenance_documents") as log_prov_documents:
        log_prov_documents.return_value = ["doc1", "doc2"]

        with patch("mlflow.log_artifact") as mlflow_log_artifact:
            logger_instance.log(item=None, identifier=None, kind="prov_documents", step=1)

            log_prov_documents.assert_called_once_with(create_graph=True, create_svg=True)
            mlflow_log_artifact.assert_any_call("doc1")
            mlflow_log_artifact.assert_any_call("doc2")


# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Anna Lappe
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Anna Lappe <anna.elisa.lappe@cern.ch> - CERN
# --------------------------------------------------------------------------------------


"""Logic to parse configuration and transform it into Ray objects."""

import logging
from typing import Dict

import ray.train
import ray.tune
from ray.tune.search.sample import Categorical, Float, Function, Integer

py_logger = logging.getLogger(__name__)


def search_space(config: Dict | None) -> Dict:
    if not config:
        py_logger.warning(
            "No search_space configuration detected. "
            "If you want to tune any hyperparameters, make sure to define them here."
        )
        return {}

    try:
        search_space = {}
        for name, param in config.items():
            if isinstance(param, (Categorical, Float, Integer, Function)):
                # The param is already a tune object and does not need to be parsed
                search_space[name] = param
                continue
            if isinstance(param, dict) and "grid_search" in param:
                # The param is already a tune grid search object and does not need to be parsed
                search_space[name] = param
                continue

            # From now on this function tries to parse the params from a dictionary
            if not isinstance(param, dict):
                raise ValueError(
                    f"Unable to parse '{param}' in the config as a tunable param."
                )

            # Convert specific keys to float if necessary
            for key in ["lower", "upper", "mean", "std"]:
                if key in param:
                    param[key] = float(param[key])

            param_type = param.pop("type")
            param = getattr(ray.tune, param_type)(**param)
            search_space[name] = param
        return search_space
    except Exception as exc:
        if hasattr(exc, "add_note"):
            # This was introduced from Python 3.11
            exc.add_note(
                f"{param} could not be set. Check that this parameter type is "
                "supported by Ray Tune at "
                "https://docs.ray.io/en/latest/tune/api/search_space.html"
            )
        raise exc



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Kalliopi Tsolaki
#
# Credit:
# - Kalliopi Tsolaki <kalliopi.tsolaki@cren.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Killian Verder <killian.verder@cern.ch> - CERN
# --------------------------------------------------------------------------------------

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import subprocess
import sys

sys.path.insert(0, os.path.abspath("../tutorials/ml-workflows/"))
sys.path.insert(0, os.path.abspath("../src"))
sys.path.insert(0, os.path.abspath("../images"))

project = "itwinai"
copyright = "2024, interTwin"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.viewcode",
    "sphinx_tabs.tabs",
    "myst_parser",
    "nbsphinx",
    "sphinx.ext.napoleon",
    "sphinx_copybutton",
    "sphinx_design",
]

source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "markdown",
    ".md": "markdown",  # ,
    # '.ipynb': 'nbsphinx'
}

templates_path = ["_templates"]
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "requirements.txt",
    "README.md",
    "testing-with-pytest.md",
    "working-with-containers.md",
]


autodoc_mock_imports = ["tensorflow", "keras"]
suppress_warnings = ["myst.xref_missing", "myst.header"]

# Enable numref
numfig = True

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output


def get_git_tag():
    try:
        return (
            subprocess.check_output(["git", "describe", "--tags", "--abbrev=0"])
            .decode("utf-8")
            .strip()
        )
    except subprocess.CalledProcessError:
        return "unknown"


# Set the version to the latest tag
version = get_git_tag()
release = version

html_theme = "sphinx_rtd_theme"  # 'alabaster'
html_static_path = ["_static"]

html_logo = "../docs/images/icon-itwinai-orange-white-subtitle.png"
html_theme_options = {"logo_only": True, "style_nav_header_background": "black"}

html_context = {"display_version": True, "release": release}

html_footer = """
<div class="custom-footer">
    <strong>Version:</strong> {{ release }}
</div>
"""

html_sidebars = {"**": [html_footer]}  # Adds the custom footer with version information



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Anna Lappe <anna.elisa.lappe@cern.ch> - CERN
# -------------------------------------------------------------------------------------

"""
``itwinai`` wrappers for well-known ML loggers.

A logger allows to save objects of different kinds:

.. list-table:: Logger kinds
   :widths: 25 25
   :header-rows: 1

   * - Object ``kind``
     - Description
   * - ``metric``
     - Number, usually representing a ML metric of interest (e.g., loss,
       accuracy).
   * - ``torch``
     - PyTorch object (e.g., tensor).
   * - ``artifact``
     - File on the local filesystem to be stored by the logger.
   * - ``figure``
     - Matplotlib of Plotly figure
   * - ``image``
     - PIL image or numpy array storing an image.
   * - ``param``
     - | Hyper-parameter (e.g., learning rate, batch size, number of layers)
       | as a primitive Python type.
   * - ``text``
     - Running text (string).
   * - ``dict``
     - Python dictionary.
   * - ``model``
     - ML model. At the moment only :class:`torch.nn.Module` is supported.
   * - ``best_model``
     - Best ML model. At the moment only :class:`torch.nn.Module` is
       supported.
   * - ``dataset``
     - Dataset object (e.g., objects of type :class:`mlflow.data.Dataset`).
   * - ``watch``
     - | WandB ``watch``: Hook into the torch model to collect gradients and
       | the topology. `More info`_.
   * - ``flops_pb``
     - Flops per batch, used by :class:`~itwinai.loggers.Prov4MLLogger`.
   * - ``flops_pb``
     - Flops per batch, used by :class:`~itwinai.loggers.Prov4MLLogger`.
   * - ``flops_pe``
     - Flops per epoch, used by :class:`~itwinai.loggers.Prov4MLLogger`.
   * - ``system``
     - System metrics, used by :class:`~itwinai.loggers.Prov4MLLogger`.
   * - ``carbon``
     - Carbon footprint information, used
       by :class:`~itwinai.loggers.Prov4MLLogger`.
   * - ``execution_time``
     - Execution time, used by :class:`~itwinai.loggers.Prov4MLLogger`.
   * - ``prov_documents``
     - Provenance documents, used by :class:`~itwinai.loggers.Prov4MLLogger`.

.. _More info:
    https://docs.wandb.ai/ref/python/watch
"""

import functools
import logging
import os
from abc import ABC, abstractmethod
from contextlib import contextmanager
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    Union,
)

if TYPE_CHECKING:
    import mlflow


BASE_EXP_NAME: str = "default_experiment"

py_logger = logging.getLogger(__name__)


class LogMixin(ABC):
    @abstractmethod
    def log(
        self,
        item: Union[Any, List[Any]],
        identifier: Union[str, List[str]],
        kind: str = "metric",
        step: Optional[int] = None,
        batch_idx: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Log ``item`` with ``identifier`` name of ``kind`` type at ``step``
        time step.

        Args:
            item (Union[Any, List[Any]]): element to be logged (e.g., metric).
            identifier (Union[str, List[str]]): unique identifier for the
                element to log(e.g., name of a metric).
            kind (str, optional): type of the item to be logged. Must be one
                among the list of self.supported_kinds. Defaults to 'metric'.
            step (Optional[int], optional): logging step. Defaults to None.
            batch_idx (Optional[int], optional): DataLoader batch counter
                (i.e., batch idx), if available. Defaults to None.
        """


def check_initialized(method: Callable) -> Callable:
    """Decorator for logger methods to check whether the logger
    was correctly initialized before calling the method."""

    @functools.wraps(method)
    def wrapper(self: "Logger", *args, **kwargs):
        if not self.is_initialized:
            raise RuntimeError(
                (
                    f"{self.__class__.__name__} has not been initialized. "
                    "Use either the ``start_logging`` context or the "
                    "``create_logger_context`` method."
                )
            )
        return method(self, *args, **kwargs)

    return wrapper


def check_not_initialized(method: Callable) -> Callable:
    """Decorator for ``create_logger_context`` method to prevent double initialization of
    a logger."""

    @functools.wraps(method)
    def wrapper(self: "Logger", *args, **kwargs):
        if self.is_initialized:
            py_logger.warning(
                f"Trying to initialize {self.__class__.__name__} twice.. "
                "Skipping initialization."
            )
            return
        return method(self, *args, **kwargs)

    return wrapper


class Logger(LogMixin):
    """Base class for logger

    Args:
        savedir (Union[Path, str], optional): filesystem location where logs are stored.
            Defaults to 'mllogs'.
        log_freq (Union[int, Literal['epoch', 'batch']], optional):
            how often should the logger fulfill calls to the `log()`
            method:

            - When set to 'epoch', the logger logs only if ``batch_idx``
              is not passed to the ``log`` method.

            - When an integer
              is given, the logger logs if ``batch_idx`` is a multiple of
              ``log_freq``.

            - When set to ``'batch'``, the logger logs always.

            Defaults to 'epoch'.
        log_on_workers (Optional[Union[int, List[int]]]): if -1, log on all
            workers; if int log on worker with rank equal to log_on_workers;
            if List[int], log on workers which rank is in the list.
            Defaults to 0 (the global rank of the main worker).
    """

    #: Location on filesystem where to store data.
    savedir: Path
    #: Supported logging 'kind's.
    supported_kinds: Tuple[str]
    #: Current worker global rank
    worker_rank: int = 0
    #: Flag to check whether the logger was correctly initialized
    is_initialized: bool = False

    _log_freq: Union[int, Literal["epoch", "batch"]]

    def __init__(
        self,
        savedir: Union[Path, str] = "mllogs",
        log_freq: Union[int, Literal["epoch", "batch"]] = "epoch",
        log_on_workers: Union[int, List[int]] = 0,
        experiment_id: Optional[str] = None,
        run_id: Optional[Union[int, str]] = None,
    ) -> None:
        self.savedir = Path(savedir)
        self.log_freq = log_freq
        self.log_on_workers = log_on_workers
        self._experiment_id = experiment_id
        self._run_id = run_id

    @property
    def experiment_id(self) -> Optional[str]:
        """Return the experiment name."""
        return self._experiment_id

    @property
    def run_id(self) -> Optional[Union[int, str]]:
        """Return the experiment version."""
        return self._run_id

    @property
    def log_freq(self) -> Union[int, Literal["epoch", "batch"]]:
        """Get ``log_feq``, namely how often should the logger
        fulfill or ignore calls to the `log()` method."""
        return self._log_freq

    @log_freq.setter
    def log_freq(self, val: Union[int, Literal["epoch", "batch"]]):
        """Sanitize log_freq value."""
        if val in ["epoch", "batch"] or (isinstance(val, int) and val > 0):
            self._log_freq = val
        else:
            raise ValueError(
                "Wrong value for 'log_freq'. Supported values are: "
                f"['epoch', 'batch'] or int > 0. Received: {val}"
            )

    @contextmanager
    def start_logging(self, rank: int = 0):
        """Start logging context.

        Args:
            rank (int): global rank of current process,
                used in distributed environments. Defaults to 0.

        Example:


        >>> with my_logger.start_logging():
        >>>     my_logger.log(123, 'value', kind='metric', step=0)


        """
        try:
            self.create_logger_context(rank=rank)
            yield
        finally:
            self.destroy_logger_context()

    @abstractmethod
    def create_logger_context(self, rank: int = 0) -> Any:
        """Initializes the logger context.

        Args:
            rank (int): global rank of current process,
                used in distributed environments. Defaults to 0.
        """

    @abstractmethod
    def destroy_logger_context(self) -> None:
        """Destroy logger."""

    @abstractmethod
    def save_hyperparameters(self, params: Dict[str, Any]) -> None:
        """Save hyperparameters.

        Args:
            params (Dict[str, Any]): hyperparameters dictionary.
        """

    def serialize(self, obj: Any, identifier: str) -> str:
        """Serializes object to disk and returns its path.

        Args:
            obj (Any): item to save.
            identifier (str): identifier of the item to log (expected to be a
                path under ``self.savedir``).

        Returns:
            str: local path of the serialized object to be logged.
        """
        import pickle

        itm_path = self.savedir / identifier
        with open(itm_path, "wb") as itm_file:
            pickle.dump(obj, itm_file)

    def should_log(self, batch_idx: Optional[int] = None) -> bool:
        """Determines whether the logger should fulfill or ignore calls to the
        `log()` method, depending on the ``log_freq`` property:

        - When ``log_freq`` is set to 'epoch', the logger logs only if
          ``batch_idx`` is not passed to the ``log`` method.

        - When ``log_freq`` is an integer
          is given, the logger logs if ``batch_idx`` is a multiple of
          ``log_freq``.

        - When ``log_freq`` is set to ``'batch'``, the logger logs always.

        It also takes into account whether logging on the current worker
        rank is allowed by ``self.log_on_workers``.

        Args:
            batch_idx (Optional[int]): the dataloader batch idx, if available.
                Defaults to None.

        Returns:
            bool: True if the logger should log, False otherwise.
        """
        # Check worker's global rank
        worker_ok = (
            isinstance(self.log_on_workers, int)
            and (self.log_on_workers == -1 or self.log_on_workers == self.worker_rank)
        ) or (
            isinstance(self.log_on_workers, list) and self.worker_rank in self.log_on_workers
        )
        if not worker_ok:
            return False

        # Check batch ID
        if batch_idx is not None:
            if isinstance(self.log_freq, int):
                if batch_idx % self.log_freq == 0:
                    return True
                return False
            if self.log_freq == "batch":
                return True
            return False
        return True


class ConsoleLogger(Logger):
    """Simplified logger.

    Args:
        savedir (Union[Path, str], optional): where to store artifacts.
            Defaults to 'mllogs'.
        log_freq (Union[int, Literal['epoch', 'batch']], optional):
            determines whether the logger should fulfill or ignore
            calls to the `log()` method. See ``Logger.should_log`` method for
            more details. Defaults to 'epoch'.
        log_on_workers (Optional[Union[int, List[int]]]): if -1, log on all
            workers; if int log on worker with rank equal to log_on_workers;
            if List[int], log on workers which rank is in the list.
            Defaults to 0 (the global rank of the main worker).
    """

    #: Supported kinds in the ``log`` method
    supported_kinds: Tuple[str] = ("torch", "artifact", "metric")

    def __init__(
        self,
        savedir: Union[Path, str] = "mllogs",
        log_freq: Union[int, Literal["epoch", "batch"]] = "epoch",
        log_on_workers: Union[int, List[int]] = 0,
    ) -> None:
        cl_savedir = Path(savedir) / "simple-logger"
        super().__init__(savedir=cl_savedir, log_freq=log_freq, log_on_workers=log_on_workers)

    @check_not_initialized
    def create_logger_context(self, rank: int = 0):
        """Initializes the logger context.

        Args:
            rank (int): global rank of current process,
                used in distributed environments. Defaults to 0.
        """
        self.worker_rank = rank

        if not self.should_log():
            self.is_initialized = True
            return

        py_logger.info(f"Initializing {self.__class__.__name__} on rank {rank}")

        if self.savedir.is_dir():
            numeric_dirs = [
                int(exp_dir.name)
                for exp_dir in self.savedir.iterdir()
                if exp_dir.is_dir() and exp_dir.name.isdigit()
            ]
            self._experiment_id = max(numeric_dirs) + 1
        else:
            self._experiment_id = 0

        self.run_path = self.savedir / str(self.experiment_id)
        self.run_path.mkdir(exist_ok=True, parents=True)

        self.is_initialized = True

    @check_initialized
    def destroy_logger_context(self):
        """Destroy logger. Do nothing."""

    @check_initialized
    def save_hyperparameters(self, params: Dict[str, Any]) -> None:
        """Save hyperparameters. Do nothing.

        Args:
            params (Dict[str, Any]): hyperparameters dictionary.
        """
        if not self.should_log():
            return

        # TODO: save hyperparams

    @check_initialized
    def log(
        self,
        item: Union[Any, List[Any]],
        identifier: Union[str, List[str]],
        kind: str = "metric",
        step: Optional[int] = None,
        batch_idx: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Print metrics to stdout and save artifacts to the filesystem.

        Args:
            item (Union[Any, List[Any]]): element to be logged (e.g., metric).
            identifier (Union[str, List[str]]): unique identifier for the
                element to log(e.g., name of a metric).
            kind (str, optional): type of the item to be logged. Must be
                one among the list of ``self.supported_kinds``.
                Defaults to 'metric'.
            step (Optional[int], optional): logging step. Defaults to None.
            batch_idx (Optional[int], optional): DataLoader batch counter
                (i.e., batch idx), if available. Defaults to None.
            kwargs: keyword arguments to pass to the logger.
        """
        if not self.should_log(batch_idx=batch_idx):
            return

        if kind == "artifact":
            import shutil

            artifact_dir = self.run_path / "artifacts" / identifier
            artifact_dir.mkdir(exist_ok=True, parents=True)

            item_path = Path(item)
            if item_path.is_file():
                target_path = artifact_dir / identifier
                shutil.copyfile(item, target_path)

            elif item_path.is_dir():
                numeric_dirs = [
                    int(exp_dir.name)
                    for exp_dir in artifact_dir.iterdir()
                    if exp_dir.is_dir() and exp_dir.name.isdigit()
                ]
                child_id = max(numeric_dirs) + 1
                target_path = artifact_dir / f"{self._experiment_id}.{child_id}"
                shutil.copytree(item, target_path, dirs_exist_ok=True)
            else:
                print(
                    f"INFO: The ConsoleLogger expects an artifact to be either a path \
                      or a directory. Received instead an item of type {type(item)}. \
                        The item will be ignored and not logged."
                )

        elif kind == "torch":
            import torch

            target_path = self.run_path / identifier
            torch.save(item, target_path)
            print(f"INFO: ConsoleLogger saved to {target_path}...")

        elif kind == "metric":
            print(f"ConsoleLogger: {identifier} = {item}")


class MLFlowLogger(Logger):
    """Abstraction around MLFlow logger.

    Args:
        savedir (Union[Path, str], optional): path on local filesystem where logs are
            stored. Defaults to 'mllogs'.
        experiment_name (str, optional): experiment name. Defaults to
            ``itwinai.loggers.BASE_EXP_NAME``.
        tracking_uri (Optional[str], optional): MLFLow tracking URI.
            Overrides ``savedir`` if given. Defaults to None.
        run_description (Optional[str], optional): run description.
            Defaults to None.
        run_name (Optional[str], optional): run name.
            Defaults to None.
        log_freq (Union[int, Literal['epoch', 'batch']], optional):
            determines whether the logger should fulfill or ignore
            calls to the `log()` method. See ``Logger.should_log`` method for
            more details. Defaults to 'epoch'.
        log_on_workers (Optional[Union[int, List[int]]]): if -1, log on all
            workers; if int log on worker with rank equal to log_on_workers;
            if List[int], log on workers which rank is in the list.
            Defaults to 0 (the global rank of the main worker).
    """

    #: Supported kinds in the ``log`` method
    supported_kinds: Tuple[str] = (
        "metric",
        "figure",
        "image",
        "artifact",
        "torch",
        "dict",
        "param",
        "text",
        "model",
        "dataset",
    )

    #: Current MLFLow experiment's run.
    active_run: "mlflow.ActiveRun"

    def __init__(
        self,
        savedir: Union[Path, str] = "mllogs",
        experiment_name: str = BASE_EXP_NAME,
        tracking_uri: Optional[str] = None,
        run_description: Optional[str] = None,
        run_name: Optional[str] = None,
        log_freq: Union[int, Literal["epoch", "batch"]] = "epoch",
        log_on_workers: Union[int, List[int]] = 0,
    ):
        mfl_savedir = Path(savedir) / "mlflow"
        super().__init__(savedir=mfl_savedir, log_freq=log_freq, log_on_workers=log_on_workers)
        self.tracking_uri = tracking_uri
        self.run_description = run_description
        self.run_name = run_name
        self.experiment_name = experiment_name

        mfl_savedir.mkdir(parents=True, exist_ok=True)

        self.tracking_uri = (
            self.tracking_uri
            or os.environ.get("MLFLOW_TRACKING_URI")
            or Path(self.savedir).resolve().as_uri()
        )
        import mlflow

        self.mlflow = mlflow

    @check_not_initialized
    def create_logger_context(self, rank: int = 0) -> "mlflow.ActiveRun":
        """Initializes the logger context. Start MLFLow run.

        Args:
            rank (int): global rank of current process,
                used in distributed environments. Defaults to 0.

        Returns:
            mlflow.ActiveRun: active MLFlow run.
        """
        self.worker_rank = rank

        if not self.should_log():
            self.is_initialized = True
            return

        py_logger.info(f"Initializing {self.__class__.__name__} on rank {rank}")

        active_run = self.mlflow.active_run()
        if active_run:
            print("Detected an active MLFlow run. Attaching to it...")
            self.active_run = active_run
        else:
            self.mlflow.set_tracking_uri(self.tracking_uri)
            self.mlflow.set_experiment(experiment_name=self.experiment_name)
            self.active_run: self.mlflow.ActiveRun = self.mlflow.start_run(
                description=self.run_description, run_name=self.run_name
            )
        self._run_id = self.active_run.info.run_id
        self._experiment_id = self.active_run.info.experiment_id

        self.is_initialized = True

        return self.active_run

    @check_initialized
    def destroy_logger_context(self):
        """Destroy logger. End current MLFlow run."""
        if not self.should_log():
            return

        self.mlflow.end_run()

    @check_initialized
    def save_hyperparameters(self, params: Dict[str, Any]) -> None:
        """Save hyperparameters as MLFlow parameters.

        Args:
            params (Dict[str, Any]): hyperparameters dictionary.
        """
        if not self.should_log():
            return

        for param_name, val in params.items():
            self.log(item=val, identifier=param_name, step=0, kind="param")

    @check_initialized
    def log(
        self,
        item: Union[Any, List[Any]],
        identifier: Union[str, List[str]],
        kind: str = "metric",
        step: Optional[int] = None,
        batch_idx: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Log with MLFlow.

        Args:
            item (Union[Any, List[Any]]): element to be logged (e.g., metric).
            identifier (Union[str, List[str]]): unique identifier for the
                element to log(e.g., name of a metric).
            kind (str, optional): type of the item to be logged. Must be
                one among the list of ``self.supported_kinds``.
                Defaults to 'metric'.
            step (Optional[int], optional): logging step. Defaults to None.
            batch_idx (Optional[int], optional): DataLoader batch counter
                (i.e., batch idx), if available. Defaults to None.
            kwargs: keyword arguments to pass to the logger.
        """
        if not self.should_log(batch_idx=batch_idx):
            return

        if kind == "metric":
            self.mlflow.log_metric(key=identifier, value=item, step=step)
        elif kind == "artifact":
            if not isinstance(item, str):
                # Save the object locally and then log it
                name = os.path.basename(identifier)
                save_path = self.savedir / ".trash" / str(name)
                save_path.mkdir(os.path.dirname(save_path), exist_ok=True)
                item = self.serialize(item, save_path)
            self.mlflow.log_artifact(local_path=item, artifact_path=identifier)
        elif kind == "model":
            import torch

            if isinstance(item, torch.nn.Module):
                self.mlflow.pytorch.log_model(item, identifier)
            else:
                print("WARNING: unrecognized model type")
        elif kind == "dataset":
            # Log mlflow dataset
            # https://mlflow.org/docs/latest/python_api/mlflow.html#mlflow.log_input
            # It may be needed to convert item into a mlflow dataset, e.g.:
            # https://mlflow.org/docs/latest/python_api/mlflow.data.html#mlflow.data.from_pandas
            # ATM delegated to the user
            if isinstance(item, self.mlflow.data.Dataset):
                self.mlflow.log_input(item)
            else:
                print("WARNING: unrecognized dataset type. Must be an MLFlow dataset")
        elif kind == "torch":
            import torch

            # Save the object locally and then log it
            name = os.path.basename(identifier)
            save_path = self.savedir / ".trash" / str(name)
            save_path.mkdir(os.path.dirname(save_path), exist_ok=True)
            torch.save(item, save_path)
            # Log into mlflow
            self.mlflow.log_artifact(local_path=save_path, artifact_path=identifier)
        elif kind == "dict":
            self.mlflow.log_dict(dictionary=item, artifact_file=identifier)
        elif kind == "figure":
            identifier = identifier if identifier.endswith(".png") else identifier + ".png"
            self.mlflow.log_figure(
                artifact_file=identifier,
                figure=item,
                save_kwargs=kwargs.get("save_kwargs"),
            )

            # # Here is a temporary patch

            # import tempfile

            # with tempfile.TemporaryDirectory() as tmpdirname:
            #     filename = (
            #         identifier if identifier.endswith(".png") else identifier + ".png"
            #     )
            #     file_path = os.path.join(tmpdirname, filename)
            #     item.savefig(file_path)
            #     self.mlflow.log_artifact(
            #         local_path=file_path,
            #         artifact_path=filename,
            #     )
        elif kind == "image":
            self.mlflow.log_image(artifact_file=identifier, image=item)
        elif kind == "param":
            self.mlflow.log_param(key=identifier, value=item)
        elif kind == "text":
            self.mlflow.log_text(artifact_file=identifier, text=item)


class WandBLogger(Logger):
    """Abstraction around WandB logger.

    Args:
        savedir (Union[Path, str], optional): location on local filesystem where logs
            are stored. Defaults to 'mllogs'.
        project_name (str, optional): experiment name. Defaults to
            ``itwinai.loggers.BASE_EXP_NAME``.
        log_freq (Union[int, Literal['epoch', 'batch']], optional):
            determines whether the logger should fulfill or ignore
            calls to the `log()` method. See ``Logger.should_log`` method for
            more details. Defaults to 'epoch'.
        log_on_workers (Optional[Union[int, List[int]]]): if -1, log on all
            workers; if int log on worker with rank equal to log_on_workers;
            if List[int], log on workers which rank is in the list.
            Defaults to 0 (the global rank of the main worker).
        offline_mode (str, optional): Use this option if working on compute
            node without internet access. Saves logs locally.
            Defaults to 'False'.
    """

    # TODO: add support for artifacts logging

    #: Supported kinds in the ``log`` method
    supported_kinds: Tuple[str] = (
        "watch",
        "metric",
        "figure",
        "image",
        "torch",
        "dict",
        "param",
        "text",
    )

    def __init__(
        self,
        savedir: Union[Path, str] = "mllogs",
        project_name: str = BASE_EXP_NAME,
        log_freq: Union[int, Literal["epoch", "batch"]] = "epoch",
        log_on_workers: Union[int, List[int]] = 0,
        offline_mode: bool = False,
    ) -> None:
        wbl_savedir = Path(savedir) / "wandb"
        super().__init__(savedir=wbl_savedir, log_freq=log_freq, log_on_workers=log_on_workers)
        self.project_name = project_name
        self.offline_mode = offline_mode

        import wandb

        self.wandb = wandb

    @check_not_initialized
    def create_logger_context(self, rank: int = 0) -> None:
        """Initializes the logger context. Init WandB run.

        Args:
            rank (int): global rank of current process,
                used in distributed environments. Defaults to 0.
        """
        self.worker_rank = rank

        if not self.should_log():
            self.is_initialized = True
            return

        py_logger.info(f"Initializing {self.__class__.__name__} on rank {rank}")

        (self.savedir / "wandb").mkdir(
            exist_ok=True,
            parents=True,
        )
        self.active_run = self.wandb.init(
            dir=self.savedir.resolve(),
            project=self.project_name,
            mode="offline" if self.offline_mode else "online",
        )
        self.is_initialized = True

    @check_initialized
    def destroy_logger_context(self):
        """Destroy logger."""
        if not self.should_log():
            return

    @check_initialized
    def save_hyperparameters(self, params: Dict[str, Any]) -> None:
        """Save hyperparameters.

        Args:
            params (Dict[str, Any]): hyperparameters dictionary.
        """
        if not self.should_log():
            return

        self.wandb.config.update(params)

    @check_initialized
    def log(
        self,
        item: Union[Any, List[Any]],
        identifier: Union[str, List[str]],
        kind: str = "metric",
        step: Optional[int] = None,
        batch_idx: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Log with WandB. Wrapper of https://docs.wandb.ai/ref/python/log

        Args:
            item (Union[Any, List[Any]]): element to be logged (e.g., metric).
            identifier (Union[str, List[str]]): unique identifier for the
                element to log(e.g., name of a metric).
            kind (str, optional): type of the item to be logged. Must be
                one among the list of ``self.supported_kinds``.
                Defaults to 'metric'.
            step (Optional[int], optional): ignored by ``WandBLogger``.
            batch_idx (Optional[int], optional): DataLoader batch counter
                (i.e., batch idx), if available. Defaults to None.
            kwargs: keyword arguments to pass to the logger.
        """
        if not self.should_log(batch_idx=batch_idx):
            return

        if kind == "watch":
            self.wandb.watch(item)
        elif kind in self.supported_kinds:
            # self.wandb.log({identifier: item}, step=step, commit=True)
            # Let WandB use its preferred step
            self.wandb.log({identifier: item}, commit=True)


class TensorBoardLogger(Logger):
    """Abstraction around TensorBoard logger, both for PyTorch and
    TensorFlow.

    Args:
        savedir (Union[Path, str], optional): location on local filesystem where logs
            are stored. Defaults to 'mllogs'.
        log_freq (Union[int, Literal['epoch', 'batch']], optional):
            determines whether the logger should fulfill or ignore
            calls to the `log()` method. See ``Logger.should_log`` method for
            more details. Defaults to 'epoch'.
        framework (Literal['tensorflow', 'pytorch'], optional):
            whether to log PyTorch or TensorFlow ML data.
            Defaults to 'pytorch'.
        log_on_workers (Optional[Union[int, List[int]]]): if -1, log on all
            workers; if int log on worker with rank equal to log_on_workers;
            if List[int], log on workers which rank is in the list.
            Defaults to 0 (the global rank of the main worker).

    Raises:
        ValueError: when ``framework`` is not recognized.
    """

    # TODO: decouple the logger into TorchTBLogger and TFTBLogger
    # and add the missing logging types supported by each.

    #: Supported kinds in the ``log`` method
    supported_kinds: Tuple[str] = ("metric", "image", "text", "figure", "torch")

    def __init__(
        self,
        savedir: Union[Path, str] = "mllogs",
        log_freq: Union[int, Literal["epoch", "batch"]] = "epoch",
        framework: Literal["tensorflow", "pytorch"] = "pytorch",
        log_on_workers: Union[int, List[int]] = 0,
    ) -> None:
        tbl_savedir = Path(savedir) / "tensorboard"
        super().__init__(savedir=tbl_savedir, log_freq=log_freq, log_on_workers=log_on_workers)
        if framework.lower() not in ["tensorflow", "pytorch"]:
            raise ValueError(
                "Accepted values for TensorBoardLogger framework are 'tensorflow' and "
                f"'pytorch'. Received {framework}"
            )
        self.framework = framework.lower()

        # if framework.lower() == "tensorflow":
        #     import tensorflow as tf

        #     self.tf = tf
        #     self.writer = tf.summary.create_file_writer(tbl_savedir.resolve().as_posix())
        # elif framework.lower() == "pytorch":
        #     from torch.utils.tensorboard import SummaryWriter

        #     self.writer = SummaryWriter(tbl_savedir.resolve().as_posix())
        # else:
        #     raise ValueError("Framework must be either 'tensorflow' or 'pytorch'")

    @check_not_initialized
    def create_logger_context(self, rank: int = 0) -> None:
        """Initializes the logger context. Init Tensorboard run.

        Args:
            rank (int): global rank of current process,
                used in distributed environments. Defaults to 0.
        """
        self.worker_rank = rank

        if not self.should_log():
            self.is_initialized = True
            return

        py_logger.info(f"Initializing {self.__class__.__name__} on rank {rank}")

        # Create the writer
        if self.framework == "tensorflow":
            import tensorflow as tf

            self.tf = tf
            self.writer = tf.summary.create_file_writer(self.savedir.resolve().as_posix())
        elif self.framework == "pytorch":
            from torch.utils.tensorboard import SummaryWriter

            self.writer = SummaryWriter(self.savedir.resolve().as_posix())
        else:
            raise ValueError("Framework must be either 'tensorflow' or 'pytorch'")

        if self.framework == "tensorflow":
            self.writer.set_as_default()

        self.is_initialized = True

    @check_initialized
    def destroy_logger_context(self):
        """Destroy logger. Close SummaryWriter."""
        if not self.should_log():
            return

        self.writer.close()

    @check_initialized
    def save_hyperparameters(self, params: Dict[str, Any]) -> None:
        """Save hyperparameters.

        Args:
            params (Dict[str, Any]): hyperparameters dictionary.
        """
        if not self.should_log():
            return

        if self.framework == "tensorflow":
            from tensorboard.plugins.hparams import api as hp

            hparams = {hp.HParam(k): v for k, v in params.items()}
            with self.writer.as_default():
                hp.hparams(hparams)
        elif self.framework == "pytorch":
            supported_params = {}

            for name, val in params.items():
                if val is None or isinstance(val, (int, float, str, bool)):
                    supported_params[name] = val
                else:
                    supported_params[name] = str(val)
                    py_logger.debug(
                        "itwinai TensorboardLogger is converting parameter "
                        f"'{name}' to string as it is of type {type(val)}. "
                        "Supported types for params are int, float, str, bool, or None."
                    )

            self.writer.add_hparams(supported_params, {})

    @check_initialized
    def log(
        self,
        item: Union[Any, List[Any]],
        identifier: Union[str, List[str]],
        kind: str = "metric",
        step: Optional[int] = None,
        batch_idx: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Log with Tensorboard.

        Args:
            item (Union[Any, List[Any]]): element to be logged (e.g., metric).
            identifier (Union[str, List[str]]): unique identifier for the
                element to log(e.g., name of a metric).
            kind (str, optional): type of the item to be logged. Must be
                one among the list of ``self.supported_kinds``.
                Defaults to 'metric'.
            step (Optional[int], optional): logging step. Defaults to None.
            batch_idx (Optional[int], optional): DataLoader batch counter
                (i.e., batch idx), if available. Defaults to None.
            kwargs: keyword arguments to pass to the logger.
        """
        if not self.should_log(batch_idx=batch_idx):
            return

        if self.framework == "tensorflow":
            with self.writer.as_default():
                if kind == "metric":
                    self.tf.summary.scalar(identifier, item, step=step)
                elif kind == "image":
                    self.tf.summary.image(identifier, item, step=step)
                elif kind == "text":
                    self.tf.summary.text(identifier, item, step=step)
                elif kind == "figure":
                    self.tf.summary.figure(identifier, item, step=step)
        elif self.framework == "pytorch":
            if kind == "metric":
                self.writer.add_scalar(identifier, item, global_step=step)
            elif kind == "image":
                self.writer.add_image(identifier, item, global_step=step)
            elif kind == "text":
                self.writer.add_text(identifier, item, global_step=step)
            elif kind == "figure":
                self.writer.add_figure(identifier, item, global_step=step)
            elif kind == "torch":
                self.writer.add_graph(item)


class LoggersCollection(Logger):
    """Wrapper of a set of loggers, allowing to use them simultaneously.

    Args:
        loggers (List[Logger]): list of itwinai loggers.
    """

    #: Supported kinds are delegated to the loggers in the collection.
    supported_kinds: Tuple[str]

    def __init__(self, loggers: List[Logger]) -> None:
        super().__init__(savedir=Path("/tmp/mllogs_LoggersCollection"), log_freq=1)
        self.loggers = loggers

    def should_log(self, batch_idx: int = None) -> bool:
        """Transparent method which delegates the `Logger.should_log``
        to individual loggers. Always returns True.

        Args:
            batch_idx (int, optional): dataloader batch index.
            Defaults to None.

        Returns:
            bool: always True.
        """
        return True

    @check_initialized
    def log(
        self,
        item: Union[Any, List[Any]],
        identifier: Union[str, List[str]],
        kind: str = "metric",
        step: Optional[int] = None,
        batch_idx: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Log on all loggers.

        Args:
            item (Union[Any, List[Any]]): element to be logged (e.g., metric).
            identifier (Union[str, List[str]]): unique identifier for the
                element to log(e.g., name of a metric).
            kind (str, optional): type of the item to be logged. Must be
                one among the list of ``self.supported_kinds``.
                Defaults to 'metric'.
            step (Optional[int], optional): logging step. Defaults to None.
            batch_idx (Optional[int], optional): DataLoader batch counter
                (i.e., batch idx), if available. Defaults to None.
            kwargs: keyword arguments to pass to the logger.
        """
        for logger in self.loggers:
            logger.log(
                item=item,
                identifier=identifier,
                kind=kind,
                step=step,
                batch_idx=batch_idx,
                **kwargs,
            )

    @check_not_initialized
    def create_logger_context(self, rank: int = 0) -> Any:
        """Initializes all loggers.

        Args:
            rank (int): global rank of current process,
                used in distributed environments. Defaults to 0.
        """
        for logger in self.loggers:
            logger.create_logger_context(rank=rank)

        self.is_initialized = True

    @check_initialized
    def destroy_logger_context(self):
        """Destroy all loggers."""
        for logger in self.loggers:
            logger.destroy_logger_context()

    @check_initialized
    def save_hyperparameters(self, params: Dict[str, Any]) -> None:
        """Save hyperparameters for all loggers.

        Args:
            params (Dict[str, Any]): hyperparameters dictionary.
        """
        for logger in self.loggers:
            logger.save_hyperparameters(params=params)


class Prov4MLLogger(Logger):
    """
    Abstraction around Prov4ML logger.

    Args:
        prov_user_namespace (str, optional): location to where provenance
            files will be uploaded. Defaults to "www.example.org".
        experiment_name (str, optional): experiment name.
            Defaults to "experiment_name".
        provenance_save_dir (Union[Path, str], optional): path where to store provenance
            files and logs. Defaults to "prov".
        save_after_n_logs (Optional[int], optional): how often to save
            logs to disk from main memory. Defaults to 100.
        create_graph (Optional[bool], optional): whether to create a
            provenance graph. Defaults to True.
        create_svg (Optional[bool], optional): whether to create an SVG
            representation of the provenance graph. Defaults to True.
        log_freq (Union[int, Literal['epoch', 'batch']], optional):
            determines whether the logger should fulfill or ignore
            calls to the `log()` method. See ``Logger.should_log`` method for
            more details. Defaults to 'epoch'.
        log_on_workers (Optional[Union[int, List[int]]]): if -1, log on all
            workers; if int log on worker with rank equal to log_on_workers;
            if List[int], log on workers which rank is in the list.
            Defaults to 0 (the global rank of the main worker).
    """

    #: Supported kinds in the ``log`` method
    supported_kinds: Tuple[str] = (
        "metric",
        "flops_pb",
        "flops_pe",
        "system",
        "carbon",
        "execution_time",
        "model",
        "best_model",
        "torch",
    )

    def __init__(
        self,
        prov_user_namespace: str = "www.example.org",
        experiment_name: str = "experiment_name",
        provenance_save_dir: Union[Path, str] = "mllogs/prov_logs",
        save_after_n_logs: Optional[int] = 100,
        create_graph: Optional[bool] = True,
        create_svg: Optional[bool] = True,
        log_freq: Union[int, Literal["epoch", "batch"]] = "epoch",
        log_on_workers: Union[int, List[int]] = 0,
    ) -> None:
        super().__init__(
            savedir=provenance_save_dir,
            log_freq=log_freq,
            log_on_workers=log_on_workers,
        )
        self.prov_user_namespace = prov_user_namespace
        self.experiment_name = experiment_name
        self.provenance_save_dir = provenance_save_dir
        self.save_after_n_logs = save_after_n_logs
        self.create_graph = create_graph
        self.create_svg = create_svg

        import mlflow
        import prov4ml

        self.prov4ml = prov4ml
        self.mlflow = mlflow

    @check_not_initialized
    def create_logger_context(self, rank: int = 0):
        """Initializes the logger context.

        Args:
            rank (int): global rank of current process,
                used in distributed environments. Defaults to 0.
        """
        self.worker_rank = rank

        if not self.should_log():
            self.is_initialized = True
            return

        py_logger.info(f"Initializing {self.__class__.__name__} on rank {rank}")

        self.prov4ml.start_run(
            prov_user_namespace=self.prov_user_namespace,
            experiment_name=self.experiment_name,
            provenance_save_dir=self.provenance_save_dir,
            save_after_n_logs=self.save_after_n_logs,
            # This class will control which workers can log
            collect_all_processes=True,
            rank=rank,
        )
        self.is_initialized = True

    @check_initialized
    def destroy_logger_context(self):
        """Destroys the logger context."""
        if not self.should_log():
            return

        self.prov4ml.end_run(create_graph=self.create_graph, create_svg=self.create_svg)

    @check_initialized
    def save_hyperparameters(self, params: Dict[str, Any]) -> None:
        if not self.should_log():
            return

        # Save hyperparams
        for param_name, val in params.items():
            self.prov4ml.log_param(key=param_name, value=val)

    @check_initialized
    def log(
        self,
        item: Union[Any, List[Any]],
        identifier: Union[str, List[str]],
        kind: str = "metric",
        step: Optional[int] = None,
        batch_idx: Optional[int] = None,
        context: Optional[str] = "training",
        **kwargs,
    ) -> None:
        """Logs with Prov4ML.

        Args:
            item (Union[Any, List[Any]]): element to be logged (e.g., metric).
            identifier (Union[str, List[str]]): unique identifier for the
                element to log(e.g., name of a metric).
            kind (str, optional): type of the item to be logged. Must be
                one among the list of ``self.supported_kinds``.
                Defaults to 'metric'.
            step (Optional[int], optional): logging step. Defaults to None.
            batch_idx (Optional[int], optional): DataLoader batch counter
                (i.e., batch idx), if available. Defaults to None.
            kwargs: keyword arguments to pass to the logger.
        """

        if not self.should_log(batch_idx=batch_idx):
            return

        if kind == "metric":
            self.prov4ml.log_metric(key=identifier, value=item, context=context, step=step)
        elif kind == "flops_pb":
            model, batch = item
            self.prov4ml.log_flops_per_batch(
                label=identifier, model=model, batch=batch, context=context, step=step
            )
        elif kind == "flops_pe":
            model, dataset = item
            self.prov4ml.log_flops_per_epoch(
                label=identifier,
                model=model,
                dataset=dataset,
                context=context,
                step=step,
            )
        elif kind == "system":
            self.prov4ml.log_system_metrics(context=context, step=step)
        elif kind == "carbon":
            self.prov4ml.log_carbon_metrics(context=context, step=step)
        elif kind == "execution_time":
            self.prov4ml.log_current_execution_time(
                label=identifier, context=context, step=step
            )
        elif kind == "model":
            self.prov4ml.save_model_version(
                model=item, model_name=identifier, context=context, step=step
            )
        elif kind == "best_model":
            self.prov4ml.log_model(
                model=item,
                model_name=identifier,
                log_model_info=True,
                log_as_artifact=True,
            )
        elif kind == "torch":
            from torch.utils.data import DataLoader

            if isinstance(item, DataLoader):
                self.prov4ml.log_dataset(dataset=item, label=identifier)
            else:
                self.prov4ml.log_param(key=identifier, value=item)
        elif kind == "prov_documents":
            prov_docs = self.prov4ml.log_provenance_documents(
                create_graph=True, create_svg=True
            )

            # Upload to MLFlow
            if self.mlflow.active_run() is not None:
                for f in prov_docs:
                    if f:
                        self.mlflow.log_artifact(f)


class EmptyLogger(Logger):
    """Dummy logger which can be used as a placeholder when a real logger is
    not available. All methods do nothing.
    """

    def __init__(
        self,
        savedir: Union[Path, str] = "mllogs",
        log_freq: int | Literal["epoch"] | Literal["batch"] = "epoch",
        log_on_workers: int | List[int] = 0,
    ) -> None:
        super().__init__(savedir, log_freq, log_on_workers)

    def create_logger_context(self, rank: int = 0):
        pass

    def destroy_logger_context(self):
        pass

    def save_hyperparameters(self, params: Dict[str, Any]) -> None:
        pass

    def log(
        self,
        item: Union[Any, List[Any]],
        identifier: Union[str, List[str]],
        kind: str = "metric",
        step: Optional[int] = None,
        batch_idx: Optional[int] = None,
        **kwargs,
    ) -> None:
        pass


class EpochTimeTracker:
    """Logger for epoch execution time during training."""

    def __init__(
        self,
        strategy_name: str,
        save_path: Path | str,
        num_nodes: int,
        should_log: bool = True,
    ) -> None:
        if isinstance(save_path, str):
            save_path = Path(save_path)

        self.should_log = should_log
        self.save_path: Path = save_path
        self.strategy_name = strategy_name
        self.num_nodes = num_nodes
        self.data = {"epoch_id": [], "time": []}

        if not self.should_log:
            print("Warning: EpochTimeLogger has been disabled!")

    def add_epoch_time(self, epoch_idx: int, time: float) -> None:
        """Add epoch time to data."""
        if not self.should_log:
            return

        self.data["epoch_id"].append(epoch_idx)
        self.data["time"].append(time)
        self.save()

    def save(self) -> None:
        """Save data to a new CSV file."""
        if not self.should_log:
            return

        import pandas as pd

        df = pd.DataFrame(self.data)
        df["name"] = self.strategy_name
        df["nodes"] = self.num_nodes

        self.save_path.parent.mkdir(parents=True, exist_ok=True)
        df.to_csv(self.save_path, index=False)
        print(f"Saving EpochTimeLogging data to '{self.save_path.resolve()}'.")



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Kalliopi Tsolaki
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Kalliopi Tsolaki <kalliopi.tsolaki@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import os
import sys
import tempfile
from typing import Any, Dict, Optional, Union

import lightning as pl
import torch
import yaml
from dataloader import ParticlesDataModule
from lightning.pytorch import Trainer as LightningTrainer
from lightning.pytorch.cli import LightningCLI
from model import ThreeDGAN
from torch import Tensor

from itwinai.components import Predictor, Trainer, monitor_exec

# from itwinai.torch.mlflow import (
#     init_lightning_mlflow,
#     teardown_lightning_mlflow
# )
from itwinai.loggers import EmptyLogger, Logger
from itwinai.serialization import ModelLoader
from itwinai.torch.inference import TorchModelLoader
from itwinai.torch.type import Batch
from itwinai.utils import load_yaml


class Lightning3DGANTrainer(Trainer):
    def __init__(self, config: Union[Dict, str], itwinai_logger: Optional[Logger] = None):
        self.save_parameters(**self.locals2params(locals()))
        super().__init__()
        if isinstance(config, str) and os.path.isfile(config):
            # Load from YAML
            config = load_yaml(config)
        self.conf = config
        self.itwinai_logger = itwinai_logger if itwinai_logger else EmptyLogger()

    @monitor_exec
    def execute(self) -> Any:
        # Parse lightning configuration
        old_argv = sys.argv
        sys.argv = ["some_script_placeholder.py"]
        cli = LightningCLI(
            args=self.conf,
            model_class=ThreeDGAN,
            datamodule_class=ParticlesDataModule,
            trainer_class=LightningTrainer,
            run=False,
            save_config_kwargs={
                "overwrite": True,
                "config_filename": "pl-training.yml",
            },
            subclass_mode_model=True,
            subclass_mode_data=True,
        )
        sys.argv = old_argv

        with self.itwinai_logger.start_logging(rank=cli.trainer.global_rank):
            # Set the logger into the LightningTrainer
            cli.trainer.itwinai_logger = self.itwinai_logger

            # Start training
            cli.trainer.fit(cli.model, datamodule=cli.datamodule)

            self._log_config(self.itwinai_logger)
            self.itwinai_logger.log(
                cli.trainer.train_dataloader, "train_dataloader", kind="torch"
            )
            self.itwinai_logger.log(
                cli.trainer.val_dataloaders, "val_dataloader", kind="torch"
            )

    def _log_config(self, logger: Logger):
        with tempfile.TemporaryDirectory(dir="/tmp") as tmp_dir:
            local_yaml_path = os.path.join(tmp_dir, "pl-conf.yaml")
            with open(local_yaml_path, "w") as outfile:
                yaml.dump(self.conf, outfile, default_flow_style=False)
            logger.log(local_yaml_path, "lightning-config", kind="artifact")


class LightningModelLoader(TorchModelLoader):
    """Loads a torch lightning model from somewhere.

    Args:
        model_uri (str): Can be a path on local filesystem
            or an mlflow 'locator' in the form:
            'mlflow+MLFLOW_TRACKING_URI+RUN_ID+ARTIFACT_PATH'
    """

    def __call__(self) -> pl.LightningModule:
        """ "Loads model from model URI.

        Raises:
            ValueError: if the model URI is not recognized
                or the model is not found.

        Returns:
            pl.LightningModule: torch lightning module.
        """
        # TODO: improve
        # # Load best model
        # loaded_model = cli.model.load_from_checkpoint(
        #     ckpt_path,
        #     lightning_conf['model']['init_args']
        # )
        return super().__call__()


class Lightning3DGANPredictor(Predictor):
    def __init__(
        self,
        model: Union[ModelLoader, pl.LightningModule],
        config: Union[Dict, str],
        name: Optional[str] = None,
    ):
        self.save_parameters(**self.locals2params(locals()))
        super().__init__(model, name)
        if isinstance(config, str) and os.path.isfile(config):
            # Load from YAML
            config = load_yaml(config)
        self.conf = config

    @monitor_exec
    def execute(
        self,
        datamodule: Optional[pl.LightningDataModule] = None,
        model: Optional[pl.LightningModule] = None,
    ) -> Dict[str, Tensor]:
        old_argv = sys.argv
        sys.argv = ["some_script_placeholder.py"]
        cli = LightningCLI(
            args=self.conf,
            model_class=ThreeDGAN,
            datamodule_class=ParticlesDataModule,
            run=False,
            save_config_kwargs={
                "overwrite": True,
                "config_filename": "pl-training.yml",
            },
            subclass_mode_model=True,
            subclass_mode_data=True,
        )
        sys.argv = old_argv

        # Override config file with inline arguments, if given
        if datamodule is None:
            datamodule = cli.datamodule
        if model is None:
            model = cli.model

        predictions = cli.trainer.predict(model, datamodule=datamodule)

        # Transpose predictions into images, energies and angles
        images = torch.cat(
            list(map(lambda pred: self.transform_predictions(pred["images"]), predictions))
        )
        energies = torch.cat(list(map(lambda pred: pred["energies"], predictions)))
        angles = torch.cat(list(map(lambda pred: pred["angles"], predictions)))

        predictions_dict = dict()
        for img, en, ang in zip(images, energies, angles):
            sample_key = f"energy={en.item()}&angle={ang.item()}"
            predictions_dict[sample_key] = img

        return predictions_dict

    def transform_predictions(self, batch: Batch) -> Batch:
        """
        Post-process the predictions of the torch model.
        """
        return batch.squeeze(1)



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

from typing import Any

from basic_components import MyDataGetter, MyDatasetSplitter, MySaver, MyTrainer

from itwinai.components import Predictor, monitor_exec
from itwinai.parser import ArgumentParser


class MyEnsemblePredictor(Predictor):
    @monitor_exec
    def execute(self, dataset, model_ensemble) -> Any:
        """
        do some predictions with model on dataset...
        """
        return dataset


if __name__ == "__main__":
    parser = ArgumentParser(description="itwinai advanced workflows tutorial")
    parser.add_argument(
        "--data-size", "-d", type=int, required=True, help="Dataset cardinality."
    )
    parser.add_argument(
        "--train-prop", type=float, required=True, help="Train split proportion."
    )
    parser.add_argument(
        "--val-prop", type=float, required=True, help="Validation split proportion."
    )
    parser.add_argument("--lr", type=float, help="Training learning rate.")
    args = parser.parse_args()

    # Save parsed arguments to configuration file.
    # Previous configurations are overwritten, which is not good,
    # but the versioning of configuration files is out of the scope
    # of this tutorial.
    parser.save(args, "advanced_tutorial_conf.yaml", format="yaml", overwrite=True)

    # Define workflow components
    getter = MyDataGetter(data_size=args.data_size)
    splitter = MyDatasetSplitter(
        train_proportion=args.train_prop,
        validation_proportion=args.val_prop,
        test_proportion=1 - args.train_prop - args.val_prop,
    )
    trainer1 = MyTrainer(lr=args.lr)
    trainer2 = MyTrainer(lr=args.lr)
    saver = MySaver()
    predictor = MyEnsemblePredictor(model=None)

    # Define ML workflow
    dataset = getter.execute()
    train_spl, val_spl, test_spl = splitter.execute(dataset)
    _, _, _, trained_model1 = trainer1.execute(train_spl, val_spl, test_spl)
    _, _, _, trained_model2 = trainer2.execute(train_spl, val_spl, test_spl)
    _ = saver.execute(trained_model1)
    predictions = predictor.execute(test_spl, [trained_model1, trained_model2])
    print()
    print("Predictions: " + str(predictions))



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# --------------------------------------------------------------------------------------

"""Scaling test of Horovod on Imagenet using Resnet."""

import os
from pathlib import Path
from timeit import default_timer as timer

import horovod.torch as hvd
import torch
import torch.optim as optim
import torchvision
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler
from utils import get_parser, imagenet_dataset, train_epoch

from itwinai.loggers import EpochTimeTracker
from itwinai.torch.reproducibility import seed_worker, set_seed


def main():
    # Parse CLI args
    parser = get_parser()
    args = parser.parse_args()

    # Check resources availability
    use_cuda = not args.no_cuda and torch.cuda.is_available()
    is_distributed = use_cuda and torch.cuda.device_count() > 0
    torch_seed = set_seed(args.rnd_seed, deterministic_cudnn=False)

    shuffle = args.shuff and args.rnd_seed is None
    persistent_workers = args.nworker > 1
    epoch_time_save_dir = Path(args.epoch_time_directory)

    train_dataset = imagenet_dataset(args.data_dir, subset_size=args.subset_size)
    train_sampler = None

    # Setting variables
    if is_distributed:
        hvd.init()

        local_rank = hvd.local_rank()
        global_rank = hvd.rank()
        global_world_size = hvd.size()

        # By default, Adasum doesn't need scaling up learning rate
        lr_scaler = hvd.size() if not args.use_adasum else 1

        # If using GPU Adasum allreduce, scale learning rate by local_size
        if args.use_adasum and hvd.nccl_built():
            lr_scaler = hvd.local_size()

        # Scale learning rate by lr_scaler
        args.lr *= lr_scaler
        pin_memory = True

        train_sampler = DistributedSampler(
            train_dataset,
            num_replicas=global_world_size,
            rank=global_rank,
            shuffle=shuffle,
        )
    else:
        # Use a single worker (either on GPU or CPU)
        local_rank = 0
        global_rank = 0
        global_world_size = 1
        pin_memory = False

    train_loader = DataLoader(
        train_dataset,
        batch_size=args.batch_size,
        sampler=train_sampler,
        num_workers=args.nworker,
        pin_memory=pin_memory,
        persistent_workers=persistent_workers,
        prefetch_factor=args.prefetch,
        generator=torch_seed,
        worker_init_fn=seed_worker,
    )

    device = torch.device(f"cuda:{local_rank}" if use_cuda else "cpu")

    # Create CNN model
    model = torchvision.models.resnet152()
    model.to(device)

    optimizer = optim.SGD(model.parameters(), lr=args.lr, momentum=args.momentum)

    if is_distributed:
        # Broadcast parameters & optimizer state
        hvd.broadcast_parameters(model.state_dict(), root_rank=0)
        hvd.broadcast_optimizer_state(optimizer, root_rank=0)

        # Compression algorithm
        compression = (
            hvd.Compression.fp16 if args.fp16_allreduce else hvd.Compression.none
        )

        # Wrap optimizer with DistributedOptimizer
        optimizer = hvd.DistributedOptimizer(
            optimizer,
            named_parameters=model.named_parameters(),
            compression=compression,
            op=hvd.Adasum if args.use_adasum else hvd.Average,
            gradient_predivide_factor=args.gradient_predivide_factor,
        )

    if global_rank == 0:
        num_nodes = os.environ.get("SLURM_NNODES", 1)
        save_path = epoch_time_save_dir / f"epochtime_horovod-bl_{num_nodes}.csv"
        epoch_time_logger = EpochTimeTracker(
            strategy_name="horovod-bl",
            save_path=save_path,
            num_nodes=int(num_nodes),
        )

    start_time = timer()
    for epoch_idx in range(1, args.epochs + 1):
        epoch_start_time = timer()

        if is_distributed:
            train_sampler.set_epoch(epoch_idx)

        train_epoch(
            model=model, device=device, train_loader=train_loader, optimizer=optimizer
        )

        # Final epoch
        if epoch_idx + 1 == args.epochs:
            train_loader.last_epoch = True

        if global_rank == 0:
            epoch_elapsed_time = timer() - epoch_start_time
            epoch_time_logger.add_epoch_time(epoch_idx, epoch_elapsed_time)
            print(f"[{epoch_idx}/{args.epochs}] - time: {epoch_elapsed_time:.2f}s")

    if global_rank == 0:
        total_time = timer() - start_time
        print(f"Training finished - took {total_time:.2f}s")
        epoch_time_logger.save()


if __name__ == "__main__":
    main()



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

"""Custom exceptions raised during sanity checks for itwinai."""


class SanityCheckError(Exception):
    """Base exception for all sanity check errors."""



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Linus Eickhoff <linus.maximilian.eickhoff@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import abc
import builtins as __builtin__
import functools
import logging
import os
import subprocess
import sys
from typing import TYPE_CHECKING, Any, Callable

from pydantic import BaseModel

if TYPE_CHECKING:
    from ray.train import ScalingConfig


py_logger = logging.getLogger(__name__)


class DistributedStrategy(abc.ABC):
    """Abstract class to define the distributed backend methods."""


class ClusterEnvironment(BaseModel):
    """Stores information about distributed environment."""

    #: Global rank of current worker, in a distributed environment.
    #: ``global_rank==0`` identifies the main worker.
    #: Defaults to 0.
    global_rank: int = 0
    #: Local rank of current worker, in a distributed environment.
    #: Defaults to 0.
    local_rank: int = 0
    #: Total number of workers in a distributed environment.
    #: Defaults to 1.
    global_world_size: int = 1
    #: Number of workers on the same node in a distributed environment.
    #: Defaults to 1.
    local_world_size: int = 1


def ray_cluster_is_running() -> bool:
    try:
        # Run the `ray status` command. It should be less overhead than ray.init()
        result = subprocess.run(
            ["ray", "status"],
            capture_output=True,
            text=True,
            check=True,
        )
        # Check if the output indicates a running cluster
        return (
            "Node status" in result.stdout
            and "Resources" in result.stdout
            and "Usage" in result.stdout
        )
    except subprocess.CalledProcessError as exc:
        # If the command fails, the cluster is not running
        py_logger.debug(
            f"Subprocess failed with return code {exc.returncode} while checking if "
            "a Ray cluster exists.\n"
            f"Stdout: {exc.stdout}\n"
            f"Stderr: {exc.stderr}"
        )
        return False
    except FileNotFoundError:
        # If `ray` command is not found, Ray is not installed
        py_logger.debug(
            "Error: 'ray' command not found while checking if a Ray cluster "
            "exists. Is Ray installed?"
        )
        return False


def detect_distributed_environment() -> ClusterEnvironment:
    """Detects distributed environment, extracting information like
    global ans local ranks, and world size.
    """
    if os.getenv("TORCHELASTIC_RUN_ID") is not None:
        # Torch elastic environment
        # https://pytorch.org/docs/stable/elastic/run.html#environment-variables
        return ClusterEnvironment(
            global_rank=os.getenv("RANK"),
            local_rank=os.getenv("LOCAL_RANK"),
            local_world_size=os.getenv("LOCAL_WORLD_SIZE"),
            global_world_size=os.getenv("WORLD_SIZE"),
        )
    # Fixes issue that OMPI_* env vars might be set despite using srun instead of mpirun
    elif int(os.getenv("OMPI_COMM_WORLD_SIZE", -1)) >= int(os.getenv("SLURM_NTASKS", 0)):
        py_logger.debug("Using Open MPI environment")
        # Open MPI environment
        # https://docs.open-mpi.org/en/v5.0.x/tuning-apps/environment-var.html
        return ClusterEnvironment(
            global_rank=os.getenv("OMPI_COMM_WORLD_RANK"),
            local_rank=os.getenv("OMPI_COMM_WORLD_LOCAL_RANK"),
            local_world_size=os.getenv("OMPI_COMM_WORLD_LOCAL_SIZE"),
            global_world_size=os.getenv("OMPI_COMM_WORLD_SIZE"),
        )
    # It is difficult to understand ranks from a Ray cluster... It could have been set up to
    # tune a model with non-distributed strategy.
    # elif ray_cluster_is_running():
    #     import ray

    #     ray_initialized_in_here = False

    #     if not ray.is_initialized():
    #         ray_initialized_in_here = True
    #         ray.init(address="auto")
    #     try:
    #         # Determine the local rank and local world size
    #         current_node = ray.util.get_node_ip_address()
    #         all_nodes = [node["NodeManagerAddress"] for node in ray.nodes()]

    #         # Filter tasks on the same node
    #         local_world_size = all_nodes.count(current_node)
    #         local_rank = (
    # all_nodes[: all_nodes.index(current_node) + 1].count(current_node) - 1
    # )
    #         cluster = ClusterEnvironment(
    #             global_rank=ray.get_runtime_context().get_node_id(),
    #             local_rank=local_rank,
    #             local_world_size=local_world_size,
    #             global_world_size=len(ray.nodes()),
    #         )
    #     finally:
    #         if ray_initialized_in_here:
    #             ray.shutdown()
    #     return cluster
    elif os.getenv("SLURM_JOB_ID") is not None:
        # https://hpcc.umd.edu/hpcc/help/slurmenv.html
        return ClusterEnvironment(
            global_rank=os.getenv("SLURM_PROCID"),
            local_rank=os.getenv("SLURM_LOCALID"),
            local_world_size=os.getenv("SLURM_NTASKS_PER_NODE", 1),
            global_world_size=os.getenv("SLURM_NTASKS"),
        )
    else:
        return ClusterEnvironment()


#: Save original builtin print before patching it in distributed environments
builtin_print = __builtin__.print


def distributed_patch_print(is_main: bool) -> Callable:
    """Disable ``print()`` when not in main worker.

    Args:
        is_main (bool): whether it is called from main worker.

    Returns:
        Callable: patched ``print()``.
    """

    def patched_print(*args, **kwself):
        """Print is disables on workers different from
        the main one, unless the print is called with
        ``force=True`` argument.
        """
        force = kwself.pop("force", False)
        if is_main or force:
            builtin_print(*args, **kwself)

    return patched_print


def suppress_workers_print(func: Callable) -> Callable:
    """Decorator to suppress ``print()`` calls in workers having global rank
    different from 0. To force printing on all workers you need to use
    ``print(..., force=True)``.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        # Disable print in workers different from the main one,
        # when in distributed environments.
        dist_grank = detect_distributed_environment().global_rank
        patched_print = distributed_patch_print(is_main=dist_grank == 0)
        previous_print_backup = __builtin__.print
        __builtin__.print = patched_print
        try:
            result = func(*args, **kwargs)
        except Exception as exc:
            # Reset print to builtin before raising the exception.
            __builtin__.print = previous_print_backup
            raise exc
        # Reset print to builtin
        __builtin__.print = previous_print_backup
        return result

    return wrapper


def suppress_workers_output(func):
    """Decorator to suppress ``stadout`` and ``stderr`` in workers having global rank
    different from 0.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # Save the original stdout and stderr
        original_stdout = sys.stdout
        original_stderr = sys.stderr

        # Get global rank
        dist_grank = detect_distributed_environment().global_rank
        try:
            if dist_grank == 0:
                # If on main worker
                return func(*args, **kwargs)

            # If not on main worker, redirect stdout and stderr to devnull
            with open(os.devnull, "w") as devnull:
                sys.stdout = devnull
                sys.stderr = devnull
                # Execute the wrapped function
                return func(*args, **kwargs)
        finally:
            # Restore original stdout and stderr
            sys.stdout = original_stdout
            sys.stderr = original_stderr

    return wrapper


def get_adaptive_ray_scaling_config() -> "ScalingConfig":
    """Returns a Ray scaling config for distributed ML training depending on the resources
    available in the Ray cluster. The number of workers is equal to the number of GPUs
    available, and if there are not GPUs two CPU-only workers are used.
    """
    import ray
    from ray.train import ScalingConfig

    # Initialize Ray if not already initialized
    if not ray.is_initialized():
        ray.init()

    # Get cluster resources
    cluster_resources = ray.cluster_resources()
    num_gpus = int(cluster_resources.get("GPU", 0))

    # Configure ScalingConfig based on GPU availability
    if num_gpus <= 1:
        # If 0 or 1 GPU, don't use GPU for training
        py_logger.debug("Returning a scaling config to run distributed ML on 2 CPUs")
        return ScalingConfig(
            num_workers=2,  # Default to 2 CPU workers
            use_gpu=False,
        )
    else:
        # If multiple GPUs, use all available GPUs
        py_logger.debug(f"Returning a scaling config to run distributed ML on {num_gpus} GPUs")
        return ScalingConfig(num_workers=num_gpus, use_gpu=True)



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

from typing import Annotated

import dagger
from dagger import Doc, dag, function, object_type


@object_type
class Singularity:
    """Manage Singularity images."""

    base_image: Annotated[dagger.Container, Doc("Base singularity image")] = (
        dag.container().from_("quay.io/singularity/docker2singularity")
    )
    docker: Annotated[
        dagger.Container | None, Doc("Docker container to convert to Singularity")
    ] = None

    @function
    def client(self) -> dagger.Container:
        """Return base image container with Singularity."""
        return self.base_image

    @function
    def convert(self) -> dagger.File:
        """Export Docker container to a Singularity file."""
        return (
            self.base_image.with_file("img.tar", self.docker.as_tarball())
            .with_exec(["singularity", "build", "img.sif", "oci-archive://img.tar"])
            .file("img.sif")
        )

    @function
    def container(self) -> dagger.Container:
        """Export Docker container to a Singularity file and return a container
        containing the generated SIF.
        """
        return self.client().with_file("container.sif", self.convert())

    @function
    async def publish(
        self,
        password: Annotated[dagger.Secret, Doc("Password for registry")],
        username: Annotated[dagger.Secret, Doc("Username for registry")],
        uri: Annotated[str, Doc("Target URI for the image")],
    ) -> str:
        """Export container and publish it to some registry."""
        print(f"The Singularity image will be published at: {uri}")
        return await (
            self.base_image.with_file("container.sif", self.convert())
            .with_secret_variable(name="SINGULARITY_DOCKER_USERNAME", secret=username)
            .with_secret_variable(name="SINGULARITY_DOCKER_PASSWORD", secret=password)
            .with_exec(["singularity", "push", "container.sif", f"{uri}"])
            .stdout()
        )

    @function
    async def remove(
        self,
        password: Annotated[dagger.Secret, Doc("Password for registry")],
        username: Annotated[dagger.Secret, Doc("Username for registry")],
        registry: Annotated[
            str,
            Doc("The registry URL where the container will be published"),
        ] = "registry.egi.eu",
        project: Annotated[
            str,
            Doc("The name of the project"),
        ] = "dev.intertwin.eu",
        name: Annotated[
            str,
            Doc("The name of the container image"),
        ] = "itwinai-dev",
        tag: Annotated[
            str,
            Doc("Tag for the container image"),
        ] = "latest",
    ) -> str:
        """Remove container from some Harbor registry."""

        cmd = (
            'DIGEST="$(curl -su $SINGULARITY_DOCKER_USERNAME:$SINGULARITY_DOCKER_PASSWORD '
            f'"https://{registry}/api/v2.0/projects/{project}/repositories/{name}/'
            f"artifacts?with_tag=true\" | jq -r '.[] | select(.tags[]?.name == \"'{tag}'\")"
            " | .digest')\""
            "; curl -u $SINGULARITY_DOCKER_USERNAME:$SINGULARITY_DOCKER_PASSWORD -X DELETE "
            f'"https://{registry}/api/v2.0/projects/{project}/repositories/{name}/'
            'artifacts/$DIGEST"'
        )

        return await (
            dag.container()
            .from_("alpine")
            .with_exec(["apk", "add", "curl", "jq"])
            .with_secret_variable(name="SINGULARITY_DOCKER_USERNAME", secret=username)
            .with_secret_variable(name="SINGULARITY_DOCKER_PASSWORD", secret=password)
            .with_exec(["sh", "-c", cmd])
            .stdout()
        )



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Jarl Sondre Sæther
#
# Credit:
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# --------------------------------------------------------------------------------------

from pathlib import Path
from typing import Set, Tuple

import pandas as pd
from scipy.constants import hour as SECONDS_IN_HOUR


def check_contains_columns(
    df: pd.DataFrame, expected_columns: Set, file_path: Path | None = None
) -> None:
    """Validates that the given DataFrame contains all the expected columns. Raises a
    ValueError if any columns are missing, including the file path in the error message
    if provided.
    """
    if not expected_columns.issubset(df.columns):
        missing_columns = expected_columns - set(df.columns)
        if file_path is not None:
            raise ValueError(
                f"Invalid data format! DataFrame at '{file_path.resolve}' is missing"
                f" some necessary columns. \nMissing columns: {missing_columns}."
            )
        else:
            raise ValueError(
                f"Invalid data format for given DataFrame. \nMissing columns:"
                f"{missing_columns}."
            )


def check_probing_interval_consistency(gpu_data_df: pd.DataFrame) -> None:
    """Checks that the probing_interval is consistent within each group of strategy
    and number of GPUs.

    Raises:
        ValueError: If the probing intervals are inconsistent for any group.
    """
    unique_intervals = gpu_data_df.groupby(["strategy", "num_global_gpus"])[
        "probing_interval"
    ].nunique()

    if (unique_intervals > 1).any():
        inconsistent_group = unique_intervals.max()
        raise ValueError(
            f"probing_interval must have the same value for each strategy and "
            f"number of GPUs, but was inconsistent for strategy: {inconsistent_group[0]} "
            f"and number of GPUs: {inconsistent_group[1]}."
        )


def calculate_gpu_statistics(
    gpu_data_df: pd.DataFrame, expected_columns: Set
) -> pd.DataFrame:
    """Calculates both the total energy expenditure (in Watt-hours) and the average GPU
    utilization for each strategy and number of GPUs. Ensures consistent probing intervals.

    Returns:
        pd.DataFrame: A DataFrame containing the total energy expenditure and
            average GPU utilization for each strategy and number of GPUs, with
            the columns ``strategy``, ``num_global_gpus``, ``total_energy_wh``,
            and ``utilization``.

    Raises:
        ValueError: If the given DataFrame does not contain the expected columns.
        ValueError: If the probing intervals are inconsistent for any group.
    """
    check_contains_columns(df=gpu_data_df, expected_columns=expected_columns)
    check_probing_interval_consistency(gpu_data_df)

    # Calculate total energy expenditure and average utilization
    gpu_data_df["energy_wh"] = (
        gpu_data_df["power"] * gpu_data_df["probing_interval"] / SECONDS_IN_HOUR
    )
    aggregated_df = (
        gpu_data_df.groupby(["strategy", "num_global_gpus"])
        .agg(
            total_energy_wh=("energy_wh", "sum"),
            utilization=("utilization", "mean"),
        )
        .reset_index()
    )

    return aggregated_df


def calculate_comp_and_comm_time(df: pd.DataFrame) -> Tuple[float, float]:
    """Calculates the time spent on computation and communication in seconds from the
    given DataFrame, assuming an NCCL backend.

    Raises:
        ValueError: If the DataFrame is missing the required columns 'name' or
        'self_cuda_time_total'.
    """
    expected_columns = {"name", "self_cuda_time_total"}
    check_contains_columns(df=df, expected_columns=expected_columns)
    comm_types = [
        "AllReduce",
        "Broadcast",
        "Reduce",
        "AllGather",
        "Gather",
        "ReduceScatter",
    ]
    nccl_comm_pattern = rf"(?:{'|'.join(comm_types)})"
    cuda_stream_pattern = r"cudaStream(?:WaitEvent|Synchronize)"

    # Any operation that is a part of PyTorch's ATen library is considered a computation
    aten_comp_pattern = r"aten::"

    comm_df = df[
        (df["name"].str.contains(nccl_comm_pattern))
        | (df["name"].str.contains(cuda_stream_pattern))
    ]
    comp_df = df[df["name"].str.contains(aten_comp_pattern)]

    comp_time = comp_df["self_cuda_time_total"].sum()
    comm_time = comm_df["self_cuda_time_total"].sum()

    # Converting from microseconds to seconds
    comp_time *= 1e-6
    comm_time *= 1e-6

    return comp_time, comm_time


def get_computation_fraction_data(df: pd.DataFrame) -> pd.DataFrame:
    """Calculates the computation fraction for each strategy and GPU configuration,
    returning a DataFrame with the results. The computation fraction is defined as the
    ratio of computation time to the total time (computation + communication).
    """
    # Sort and create cartesian product of unique strategies and GPU counts
    unique_num_gpus = sorted(df["num_gpus"].unique(), key=lambda x: int(x))
    unique_strategies = sorted(df["strategy"].unique())
    index = pd.MultiIndex.from_product(
        [unique_strategies, unique_num_gpus], names=["strategy", "num_gpus"]
    )

    # Group by strategy and number of GPUs, calculate computation fraction
    def compute_fraction(group):
        comp_time, comm_time = calculate_comp_and_comm_time(df=group)
        return comp_time / (comp_time + comm_time + 1e-10)

    grouped = df.groupby(["strategy", "num_gpus"]).apply(compute_fraction)

    # Reindex to fill in missing combinations with NaN
    result_df = pd.DataFrame(grouped.reindex(index))
    result_df = result_df.reset_index()
    result_df.columns = ["strategy", "num_gpus", "computation_fraction"]
    return result_df



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

from .dummy_components import (
    FakeGetter,
    FakeGetterExec,
    FakePreproc,
    FakePreprocExec,
    FakeSaver,
    FakeSaverExec,
    FakeSplitter,
    FakeSplitterExec,
    FakeTrainer,
    FakeTrainerExec,
)

_ = (
    FakeGetter,
    FakeGetterExec,
    FakePreproc,
    FakePreprocExec,
    FakeSaver,
    FakeSaverExec,
    FakeSplitter,
    FakeSplitterExec,
    FakeTrainer,
    FakeTrainerExec,
)



"""Test Ray Tune integration in the itwinai TorchTrainer."""

import time
from types import MethodType

import numpy as np
import pytest
import ray
from ray import tune
from ray.train import RunConfig, ScalingConfig
from ray.tune import TuneConfig
from ray.tune.schedulers import AsyncHyperBandScheduler

from itwinai.distributed import ray_cluster_is_running
from itwinai.torch.trainer import TorchTrainer


@pytest.mark.hpc
@pytest.mark.ray_dist
def test_tuning_mnist_ray(mnist_datasets, shared_tmp_path, mnist_net):
    """Test HPO for TorchTrainer on MNIST."""

    assert ray_cluster_is_running(), "Ray cluster not detected. Aborting tests"

    def dummy_train(self: TorchTrainer):
        """Dummy training function"""

        # Check that the strategy is NOT running distributed ML
        assert self.strategy.global_rank() == 0, (
            "Global rank is not 0. Does the strategy think it's distributed?"
        )
        assert self.strategy.global_world_size() == 1, (
            "Global world size is not 1. Does the strategy think it's distributed?"
        )
        # assert isinstance(self.strategy, NonDistributedStrategy), (
        #     "training should not be distributed"
        # )

        # Check that the hyperparameters are used
        assert self.config.batch_size != INITIAL_BATCH_SIZE, "Batch size was not overridden"

        for i in range(1, self.epochs + 1):
            time.sleep(0.1)
            # Simulate the reporting of validation loss at the end of an epoch
            self.ray_report(
                metrics={"loss": 2 / i + np.random.randn()},
            )

    INITIAL_BATCH_SIZE = 17
    search_space = {"batch_size": tune.qrandint(128, 1024, 128)}
    tune_config = TuneConfig(
        metric="loss",
        mode="min",
        search_alg=None,  # Random search
        scheduler=AsyncHyperBandScheduler(max_t=10, grace_period=3, reduction_factor=4),
        num_samples=2,
    )
    ray_run_config = RunConfig(storage_path=shared_tmp_path / "ray_checkpoints")
    scaling_config = ScalingConfig(
        num_workers=1,
        use_gpu=True,
        resources_per_worker={"GPU": 1, "CPU": 8},
    )

    ckpt_path = shared_tmp_path / "my_checkpoints"
    training_config = dict(optimizer="sgd", loss="nllloss", batch_size=INITIAL_BATCH_SIZE)
    trainer = TorchTrainer(
        model=mnist_net,
        config=training_config,
        epochs=50,
        # The strategy should fallback to non-distributed strategy if
        # not enough resources are granted to each trial
        strategy="ddp",
        ray_scaling_config=scaling_config,
        ray_run_config=ray_run_config,
        ray_search_space=search_space,
        ray_tune_config=tune_config,
        checkpoint_every=1,
        checkpoints_location=ckpt_path,
    )

    train_set, val_set = mnist_datasets

    # Monkey patch train function
    trainer.train = MethodType(dummy_train, trainer)

    # Train
    # TODO: prevent strategy cleanup?
    trainer.execute(train_set, val_set)

    # Check if at least one trial failed
    for result in trainer.tune_result_grid:
        assert not result.error, str(result.error)


@pytest.mark.hpc
@pytest.mark.ray_dist
def test_tuning_dist_ml_mnist_ray(mnist_datasets, shared_tmp_path, mnist_net):
    """Test HPO + distributed ML for each trial, for TorchTrainer on MNIST.
    Note that this test requires at least 4 GPUs to be executed.
    """

    num_gpus = int(ray.cluster_resources().get("GPU", 0))
    num_trials = 2
    gpus_per_trial = num_gpus // num_trials

    # TODO: remove this condition in future PRs!
    if gpus_per_trial <= 1:
        return

    assert gpus_per_trial > 1, "Not enough resources to run distributed ML under HPO."

    def dummy_train(self: TorchTrainer):
        """Dummy training function"""

        # Check that the strategy is running distributed ML
        assert self.strategy.global_world_size() == gpus_per_trial, (
            "Global world size is not 1. Does the strategy think it's distributed?"
        )

        # Check that the hyperparameters are used
        assert self.config.batch_size != INITIAL_BATCH_SIZE, "Batch size was not overridden"

        for i in range(1, self.epochs + 1):
            time.sleep(0.1)
            # Simulate the reporting of validation loss at the end of an epoch
            self.ray_report(
                metrics={"loss": 2 / i + np.random.randn()},
            )

    INITIAL_BATCH_SIZE = 17
    search_space = {"batch_size": tune.choice([32, 50, 64])}
    tune_config = TuneConfig(
        metric="loss",
        mode="min",
        search_alg=None,  # Random search
        scheduler=AsyncHyperBandScheduler(max_t=10, grace_period=3, reduction_factor=4),
        num_samples=num_trials,
    )
    ray_run_config = RunConfig(storage_path=shared_tmp_path / "ray_checkpoints")
    scaling_config = ScalingConfig(
        num_workers=gpus_per_trial,
        use_gpu=True,
        resources_per_worker={"GPU": gpus_per_trial, "CPU": 8},
    )

    ckpt_path = shared_tmp_path / "my_checkpoints"
    training_config = dict(optimizer="sgd", loss="nllloss", batch_size=INITIAL_BATCH_SIZE)
    trainer = TorchTrainer(
        model=mnist_net,
        config=training_config,
        epochs=50,
        # The strategy should fallback to non-distributed strategy if
        # not enough resources are granted to each trial
        strategy="ddp",
        ray_scaling_config=scaling_config,
        ray_run_config=ray_run_config,
        ray_search_space=search_space,
        ray_tune_config=tune_config,
        checkpoint_every=1,
        checkpoints_location=ckpt_path,
    )

    train_set, val_set = mnist_datasets

    # Monkey patch train function
    trainer.train = MethodType(dummy_train, trainer)

    # Train
    # TODO: prevent strategy cleanup?
    trainer.execute(train_set, val_set)

    # Check if at least one trial failed
    for result in trainer.tune_result_grid:
        assert not result.error, str(result.error)



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Alex Krochak
#
# Credit:
# - Alex Krochak <o.krochak@fz-juelich.de> - FZJ
# --------------------------------------------------------------------------------------

"""Tests for radio-astronomy use case.

Intended to be integration tests, to make sure that updates in the code base
do not break use cases' workflows.

This is meant to be run from the main itwinai directory, not the use-case folder !!!
"pytest use-cases/radio-astronomy/tests/test_radio-astronomy.py"

NOTE FOR DEVELOPERS: if you are editing this file, make sure that entries in 
use-cases/radio-astronomy/.config-test.yaml are updated accordingly !!! 
"""

import os
import subprocess
from pathlib import Path
import shutil

import pytest

USECASE_FOLDER = Path("use-cases", "radio-astronomy").resolve()

@pytest.fixture
def torch_env() -> str:
    """Returns absolute path to torch virtual environment."""
    env_path = Path(os.environ.get("TORCH_ENV", "./.venv"))
    return str(env_path.resolve())

@pytest.fixture
def syndata(tmp_path, torch_env,install_requirements):
    # This fixture implicitly tests the synthetic data generation pipeline
    install_requirements(USECASE_FOLDER, torch_env)

    cmd_data = (
        f"{torch_env}/bin/itwinai exec-pipeline --config-name .config-test "
        f"+pipe_key=syndata_pipeline ++syndata_test_dir={tmp_path}/ "
    )
    if len(os.listdir(tmp_path)) == 0:  # only run if directory is empty
        # Copy the necessary files to the temporary directory for testing
        shutil.copy(USECASE_FOLDER / ".config-test.yaml", tmp_path)
        shutil.copy(USECASE_FOLDER / "data.py", tmp_path)
        shutil.copy(USECASE_FOLDER / "trainer.py", tmp_path)

        subprocess.run(cmd_data.split(), check=True, cwd=tmp_path)

    return tmp_path

@pytest.fixture
def generate_unet(torch_env, syndata):
    """Generate the U-Net model for the Filter-CNN test. """
    cmd = (
        f"{torch_env}/bin/itwinai exec-pipeline --config-name .config-test "
        f"+pipe_key=unet_pipeline ++image_directory={syndata}/ ++mask_directory={syndata}/ "
    )

    subprocess.run(cmd.split(), check=True, cwd=syndata)

# @pytest.mark.skip(reason="dependent on .test_dataset, incoroporated into integration test")
@pytest.mark.functional
def test_radio_astronomy_unet(torch_env, syndata, install_requirements):
    """Test U-Net Pulsar-DDT trainer by running it end-to-end
    via the config-test.yaml configuration file."""

    install_requirements(USECASE_FOLDER, torch_env)

    cmd = (
        f"{torch_env}/bin/itwinai exec-pipeline --config-name .config-test "
        f"+pipe_key=unet_pipeline ++image_directory={syndata}/ ++mask_directory={syndata}/ "
    )

    subprocess.run(cmd.split(), check=True, cwd=syndata)

@pytest.mark.functional
def test_radio_astronomy_filtercnn(torch_env, syndata, generate_unet, install_requirements):
    """Test Filter-CNN Pulsar-DDT trainer by running it end-to-end
    via the config-test.yaml configuration file. Requires the U-Net model to be present."""

    install_requirements(USECASE_FOLDER, torch_env)

    cmd = (
        f"{torch_env}/bin/itwinai exec-pipeline --config-name .config-test "
        f"+pipe_key=fcnn_pipeline ++image_directory={syndata}/ ++mask_directory={syndata}/ "
    )

    subprocess.run(cmd.split(), check=True, cwd=syndata)

@pytest.mark.functional
def test_radio_astronomy_cnn1d(torch_env, syndata, install_requirements):
    """Test CNN-1D Pulsar-DDT trainer by running it end-to-end
    via the config-test.yaml configuration file."""

    install_requirements(USECASE_FOLDER, torch_env)

    cmd = (
        f"{torch_env}/bin/itwinai exec-pipeline --config-name .config-test "
        f"+pipe_key=cnn1d_pipeline ++image_directory={syndata}/ ++mask_directory={syndata}/ "
    )

    subprocess.run(cmd.split(), check=True, cwd=syndata)

@pytest.mark.skip(reason="dependent on large real data set")
@pytest.mark.functional
def test_radio_astronomy_evaluate(torch_env):
    """Test the evaluate pipeline by running it end-to-end
    via the config-test.yaml configuration file."""

    cmd = (
        f"{torch_env}/bin/itwinai exec-pipeline "
        f"--config-name .config-test "
        f"+pipe_key=evaluate_pipeline "
    )

    ## Run the pipeline and check file generation in the use-case folder
    subprocess.run(cmd.split(), check=True, cwd=USECASE_FOLDER)
    ## Clean up the use-case folder
    subprocess.run("./.pytest-clean", check=True, cwd=USECASE_FOLDER)



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Jarl Sondre Sæther
#
# Credit:
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import functools
import time
from multiprocessing import Manager, Process
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, List

import pandas as pd

from .backend import init_backend

if TYPE_CHECKING:
    from itwinai.torch.trainer import TorchTrainer

logging_columns = [
    "sample_idx",
    "utilization",
    "power",
    "local_rank",
    "node_idx",
    "num_global_gpus",
    "strategy",
    "probing_interval",
]


def probe_gpu_utilization_loop(
    node_idx: int,
    num_local_gpus: int,
    num_global_gpus: int,
    strategy_name: str,
    log_dict: Any,
    stop_flag: Any,
    probing_interval: int = 2,
    warmup_time: int = 5,
) -> None:
    """Logs the GPU utilization across all availble GPUs on a single node. Is meant to
    be called by multiprocessing's Process and expects variables to be shared using
    a multiprocessing.Manager object. Logs utilization into `log_dict` until
    stop_flag.value is set to True.

    Args:
        node_idx: The index of the compute node that the function is called by, used
            for logging purposes.
        num_local_gpus: Number of GPUs on the current compute node.
        num_global_gpus: Number of GPUs on all nodes combined.
        strategy: Which distributed strategy is being used, e.g. "ddp" or "horovod".
        log_dict: Dictionary for storing logging data on. Should be managed by a
            multiprocessing.Manager object.
        stop_flag: Shared value telling the function when to stop logging. Should be
            managed by a multiprocessing.Manager object.
        probing_interval: How long to wait between each time a read of the GPU
            utilization is done.
        warmup_time: How long to wait before logging starts, allowing the training to
            properly start before reading.

    """
    if not set(logging_columns).issubset(set(log_dict.keys())):
        missing_columns = set(logging_columns) - set(log_dict.keys())
        raise ValueError(f"log_dict is missing the following columns: {missing_columns}")

    # load management library backend
    man_lib_type, handles, man_lib = init_backend()
    # ensure all gpu handles where retrieved
    if len(handles) != num_local_gpus:
        raise ValueError(f"Expected {num_local_gpus} handles, but got {len(handles)}.")

    time.sleep(warmup_time)

    sample_idx = 0
    while not stop_flag.value:
        for idx in range(num_local_gpus):
            if man_lib_type == "nvidia":
                handle = handles[idx]
                gpu_util = man_lib.nvmlDeviceGetUtilizationRates(handle).gpu
                power = man_lib.nvmlDeviceGetPowerUsage(handle) / 1000.0  # mW -> W

            elif man_lib_type == "amd":
                handle = handles[idx]
                gpu_util = man_lib.amdsmi_get_gpu_activity(handle)["gfx_activity"]  # W
                power = man_lib.amdsmi_get_power_info(handle)["average_socket_power"]

            log_dict["sample_idx"].append(sample_idx)
            log_dict["utilization"].append(gpu_util)
            log_dict["power"].append(power)
            log_dict["local_rank"].append(idx)
            log_dict["node_idx"].append(node_idx)
            log_dict["num_global_gpus"].append(num_global_gpus)
            log_dict["strategy"].append(strategy_name)
            log_dict["probing_interval"].append(probing_interval)

        sample_idx += 1
        time.sleep(probing_interval)


def measure_gpu_utilization(method: Callable) -> Callable:
    """Decorator for measuring GPU utilization and storing it to a .csv file."""

    def write_logs_to_file(utilization_logs: List[Dict], output_path: Path) -> None:
        dataframes = []
        for log in utilization_logs:
            if len(log) == 0:
                continue
            dataframes.append(pd.DataFrame(log))

        log_df = pd.concat(dataframes)
        log_df.to_csv(output_path, index=False)
        print(f"Writing GPU energy dataframe to '{output_path.resolve()}'.")

    @functools.wraps(method)
    def measured_method(self: "TorchTrainer", *args, **kwargs) -> Any:
        if not self.measure_gpu_data:
            print("Warning: Profiling of GPU data has been disabled!")
            return method(self, *args, **kwargs)

        gpu_probing_interval = 1
        warmup_time = 5

        strategy = self.strategy
        strategy_name = strategy.name

        local_rank = strategy.local_rank()
        global_rank = strategy.global_rank()
        num_global_gpus = strategy.global_world_size()
        num_local_gpus = strategy.local_world_size()
        node_idx = global_rank // num_local_gpus

        gpu_monitor_process = None
        manager = None
        stop_flag = None
        data = None

        # Starting a child process once per node
        if local_rank == 0:
            # Setting up shared variables for the child process
            manager = Manager()
            data = manager.dict()
            for col in logging_columns:
                data[col] = manager.list()
            stop_flag = manager.Value("i", False)

            gpu_monitor_process = Process(
                target=probe_gpu_utilization_loop,
                kwargs={
                    "node_idx": node_idx,
                    "num_local_gpus": num_local_gpus,
                    "num_global_gpus": num_global_gpus,
                    "strategy_name": strategy_name,
                    "log_dict": data,
                    "stop_flag": stop_flag,
                    "probing_interval": gpu_probing_interval,
                    "warmup_time": warmup_time,
                },
            )
            gpu_monitor_process.start()

        local_utilization_log = {}
        try:
            result = method(self, *args, **kwargs)
        finally:
            if local_rank == 0:
                stop_flag.value = True
                grace_period = 5  # extra time to let process finish gracefully
                gpu_monitor_process.join(timeout=gpu_probing_interval + grace_period)

                # Converting the shared log to non-shared log
                local_utilization_log = {key: list(data[key]) for key in data.keys()}
                manager.shutdown()

        global_utilization_log = strategy.gather_obj(local_utilization_log, dst_rank=0)
        if strategy.is_main_worker:
            output_dir = Path(f"scalability-metrics/{self.run_id}/gpu-energy-data")
            output_dir.mkdir(exist_ok=True, parents=True)
            output_path = output_dir / f"{strategy_name}_{num_global_gpus}.csv"

            write_logs_to_file(global_utilization_log, output_path)

        return result

    return measured_method
